import momentsml.sim
import numpy as np
import random # np.random.choice is only available for newer numpys...

import itertools


def contracted_rayleigh(sigma, max_val, p):
	"""
	A distribution with finite support.
	"""
	tmp = np.random.rayleigh(sigma)
	return (tmp / np.power(1 + np.power(tmp / max_val, p), 1.0 / p))




class Simple1(momentsml.sim.params.Params):
	"""
	Sersics with shear.
	Single PSF.
	"""
	
	
	def __init__(self):
		momentsml.sim.params.Params.__init__(self)
		self.snc_type = 100
		self.sr = 0.10 # shear range

	
	def stat(self):
		"""
		stat: called for each catalog (stat is for stationnary)
		"""
		return {"snc_type":self.snc_type}
	
	
	def draw(self, ix, iy, nx, ny):
		"""
		draw: called for each galaxy	
		"""
	
		########## Shear #########
		#tru_s = contracted_rayleigh(0.01, 0.9, 4)
		#tru_s_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		#(tru_s1, tru_s2) = (tru_s * np.cos(2.0 * tru_s_theta), tru_s * np.sin(2.0 * tru_s_theta))
		
		tru_s1 = np.random.uniform(-self.sr, self.sr)
		tru_s2 = np.random.uniform(-self.sr, self.sr)	
		tru_mu = 1.0

		########## PSF ##########
		
		tru_psf_sigma = 1.5
		tru_psf_g1 = 0.0
		tru_psf_g2 = 0.0
	
		########## Galaxy ##########
		
		tru_type = 1 # Sersic	
		#tru_flux = np.random.uniform(1000.0, 10000.0)
		#tru_sigma = np.random.uniform(2.0, 10.0)
		
		tru_flux = 20.0 + 100.0 * (1.0 - (float(iy)/(float(ny)-1.0)))
		tru_rad = np.random.uniform(3.0, 5.0)
		tru_sersicn = 2.0
		
		tru_g = contracted_rayleigh(0.25, 0.6, 4)
		tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		#tru_g1 = np.random.uniform(-0.3, 0.3)
		#tru_g2 = np.random.uniform(-0.3, 0.3)	
		tru_g = np.hypot(tru_g1, tru_g2)
			
		########## Noise ##########

		tru_sky_level = 0.0
		tru_gain = -1.0
		tru_read_noise = 0.05
		
		return { # It's ugly to not directly fill this dict, but this makes it clearer what is actually returned:
					
			"tru_type":tru_type,
			"tru_flux":tru_flux,
			"tru_rad":tru_rad,
			"tru_sersicn":tru_sersicn,
			"tru_g1":tru_g1,
			"tru_g2":tru_g2,
			"tru_g":tru_g, # only useful for some plots
			
			"tru_sky_level":tru_sky_level, # in ADU, just for generating noise, will not remain in the image
			"tru_gain":tru_gain, # in photons/ADU. Make this negative to have no Poisson noise
			"tru_read_noise":tru_read_noise, # in photons if gain > 0.0, otherwise in ADU.Set this to zero to have no flat Gaussian noise
			
			"tru_s1":tru_s1,
			"tru_s2":tru_s2,
			"tru_mu":tru_mu,
			
			"tru_psf_sigma":tru_psf_sigma,
			"tru_psf_g1":tru_psf_g1,
			"tru_psf_g2":tru_psf_g2,
			
		}




class Simple2(momentsml.sim.params.Params):
	"""
	Sersics with shear.
	Single PSF.
	# This one is more random in flux, we don't draw sorted or gridded values.
	"""
	
	
	def __init__(self):
		momentsml.sim.params.Params.__init__(self)
		self.snc_type = 100
		self.sr = 0.10 # shear range

	
	def stat(self):
		"""
		stat: called for each catalog (stat is for stationnary)
		"""
		return {"snc_type":self.snc_type}
	
	
	def draw(self, ix, iy, nx, ny):
		"""
		draw: called for each galaxy	
		"""
	
		########## Shear #########
		#tru_s = contracted_rayleigh(0.01, 0.9, 4)
		#tru_s_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		#(tru_s1, tru_s2) = (tru_s * np.cos(2.0 * tru_s_theta), tru_s * np.sin(2.0 * tru_s_theta))
		
		tru_s1 = np.random.uniform(-self.sr, self.sr)
		tru_s2 = np.random.uniform(-self.sr, self.sr)	
		tru_mu = 1.0

		########## PSF ##########
		
		tru_psf_sigma = 1.5
		tru_psf_g1 = 0.0
		tru_psf_g2 = 0.0
	
		########## Galaxy ##########
		
		tru_type = 1 # Sersic	
		#tru_flux = np.random.uniform(1000.0, 10000.0)
		#tru_sigma = np.random.uniform(2.0, 10.0)
		
		tru_flux = np.random.uniform(20.0, 120.0)
		tru_rad = np.random.uniform(3.0, 10.0)
		tru_sersicn = np.random.uniform(1.0, 4.0)
		
		tru_g = contracted_rayleigh(0.25, 0.6, 4)
		tru_theta = 2.0 * np.pi * np.random.uniform(0.0, 1.0)		
		(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		#tru_g1 = np.random.uniform(-0.3, 0.3)
		#tru_g2 = np.random.uniform(-0.3, 0.3)	
		tru_g = np.hypot(tru_g1, tru_g2)
			
		########## Noise ##########

		tru_sky_level = 0.0
		tru_gain = -1.0
		tru_read_noise = 0.05
		
		return { # It's ugly to not directly fill this dict, but this makes it clearer what is actually returned:
					
			"tru_type":tru_type,
			"tru_flux":tru_flux,
			"tru_rad":tru_rad,
			"tru_sersicn":tru_sersicn,
			"tru_g1":tru_g1,
			"tru_g2":tru_g2,
			"tru_g":tru_g, # only useful for some plots
			
			"tru_sky_level":tru_sky_level, # in ADU, just for generating noise, will not remain in the image
			"tru_gain":tru_gain, # in photons/ADU. Make this negative to have no Poisson noise
			"tru_read_noise":tru_read_noise, # in photons if gain > 0.0, otherwise in ADU.Set this to zero to have no flat Gaussian noise
			
			"tru_s1":tru_s1,
			"tru_s2":tru_s2,
			"tru_mu":tru_mu,
			
			"tru_psf_sigma":tru_psf_sigma,
			"tru_psf_g1":tru_psf_g1,
			"tru_psf_g2":tru_psf_g2,
			
		}




import numpy as np
import tenbilac

import logging
logging.basicConfig(level=logging.INFO)



net = tenbilac.net.Net(ni=2, nhs=[3, 3], onlyid=True)

params = net.get_params_ref()

print params

net.addnoise()

print params


net.setidentity()

print params

print net.report()

ni = 2
ngal = 500
nrea = 10
no = 1


#inputs = np.random.randn(ni * ngal).reshape((ni, ngal))
inputs = np.ones(ni * ngal * nrea).reshape((nrea, ni, ngal))


#inputs = np.array([[1, 1, -1], [2, 2, -2]])
#inputs = np.array([1, 2])


print "inputs shape", inputs.shape

params[0] = 10
params[9] = 20
params[21] = 20

print net.report()

outputs = net.run(inputs)
print "ouputs shape ", outputs.shape



targets = np.ones(no * ngal).reshape((no, ngal))

# Prepare a traindata object 
dat = tenbilac.data.Traindata(inputs=inputs, targets=targets)

# Generate the network
training = tenbilac.train.Training(net, dat, errfctname="msb")

# We train this normal (non-inverse) regression with params as inputs, and observations as output:
training.opt(algo="bfgs", maxiter=3)



print net.report()



"""
General helpers
"""

import numpy as np
import os
import cPickle as pickle
import gzip


import logging
logger = logging.getLogger(__name__)



def writepickle(obj, filepath, protocol = -1):
	"""
	I write your python object obj into a pickle file at filepath.
	If filepath ends with .gz, I'll use gzip to compress the pickle.
	Leave protocol = -1 : I'll use the latest binary protocol of pickle.
	"""
	if os.path.splitext(filepath)[1] == ".gz":
		pkl_file = gzip.open(filepath, 'wb')
	else:
		pkl_file = open(filepath, 'wb')
	
	pickle.dump(obj, pkl_file, protocol)
	pkl_file.close()
	logger.info("Wrote %s" % filepath)
	
	
def readpickle(filepath):
	"""
	I read a pickle file and return whatever object it contains.
	If the filepath ends with .gz, I'll unzip the pickle file.
	"""
	if os.path.splitext(filepath)[1] == ".gz":
		pkl_file = gzip.open(filepath,'rb')
	else:
		pkl_file = open(filepath, 'rb')
	obj = pickle.load(pkl_file)
	pkl_file.close()
	logger.info("Read %s" % filepath)
	return obj

def sigma_clip_plus(arr, maxdev, get_indices=False):
	"""
	Removes iteratively all points that are upwards of 

		mean(arr) + maxdev * std(arr)

	:param arr: a numpy array (1D) containing the data points.
	:param maxdev: maximum allowed deviation
	:param get_indices: if True returns arr, keys otherwise just keys
	"""

	lenp = 1
	lena = 0

	keys = np.arange(len(arr))
	arr = np.asarray(arr)

	while lenp > lena:
		_thr = np.mean(arr) + maxdev * np.std(arr)
		lenp = np.size(arr)
		sel = arr < _thr
		keys = keys[sel]
		arr = arr[sel]
		lena = np.size(arr)
		
		if lena == 1:
			break

	if get_indices:
		return arr, keys
	else:
		return arr



import momentsml.sim
import numpy as np
import random # np.random.choice is only available for newer numpys...

import itertools





class Simple1(momentsml.sim.params.Params):
	"""
	No PSF, just round Gaussians
	"""
	
	
	def __init__(self):
		momentsml.sim.params.Params.__init__(self)

	
	def stat(self):
		"""
		stat: called for each catalog (stat is for stationnary)
		"""
		return {"snc_type":1}
	
	
	def draw(self, ix, iy, nx, ny):
		"""
		draw: called for each galaxy	
		"""
		
		
		gain = 3.1 # electrons/ADU
		ron = 4.2 # electrons
		skyback = 22.35 # mag per arcsec2, dominated by zodiacal light
		#zeropoint = 25.9 # mag. Should give SNR 10 when observing with 3 x 565 second exposures.
		
		zeropoint = 25.9
		
		exptime = 3.0*565.0	# seconds
		
		#########  No Lensing

		tru_sky_level = 0.01 * exptime * 10**((skyback - zeropoint)/(-2.5))  # In ADU per pixel. 0.01 because of the pixel size of 0.1 arcsec. No gain, as in ADU!
		tru_gain = gain
		tru_read_noise = ron
		
		#########  No Lensing
		
		tru_s1 = 0.0
		tru_s2 = 0.0
		tru_mu = 1.0



		
		########## Galaxy ##########
		
		tru_type = 1 # 0 Gaussian, 1 sersic
		tru_sersicn = 2.0
		
		
		#tru_mag = 24.5
		tru_mag = 23.0 + float(iy)/float(ny) * 2.5

		tru_flux = (exptime / gain) * 10**((tru_mag - zeropoint)/(-2.5))

		
		#factor = 3.2 # for serscin 4
		factor = 1.4 # for serscin 2
		#factor = 1.0 # for sersic 1
		tru_rad = factor * 4.3/2.0
		
		tru_g = 0.0
		tru_theta = 0.0	
		(tru_g1, tru_g2) = (tru_g * np.cos(2.0 * tru_theta), tru_g * np.sin(2.0 * tru_theta))
		
		tru_g = np.hypot(tru_g1, tru_g2)
			
		
		
		return { # It's ugly to not directly fill this dict, but this makes it clearer what is actually returned:
					
			"tru_type":tru_type,
			"tru_flux":tru_flux,
			"tru_rad":tru_rad,
			"tru_mag":tru_mag,
			"zeropoint":zeropoint,
			#"tru_sigma":tru_sigma,
			"tru_sersicn":tru_sersicn,
			"tru_g1":tru_g1,
			"tru_g2":tru_g2,
			"tru_g":tru_g, # only useful for some plots
			
			"tru_sky_level":tru_sky_level, # in ADU, just for generating noise, will not remain in the image
			"tru_gain":tru_gain, # in photons/ADU. Make this negative to have no Poisson noise
			"tru_read_noise":tru_read_noise, # in photons if gain > 0.0, otherwise in ADU.Set this to zero to have no flat Gaussian noise
			
			"tru_s1":tru_s1,
			"tru_s2":tru_s2,
			"tru_mu":tru_mu,
			
			"tru_psf_sigma":0.75,
			"tru_psf_g1":0.0,
			"tru_psf_g2":0.0,

			
		}





"""
A module to hold some maths or statistics oriented functions woring on numpy arrays.
"""

import numpy as np
import scipy.stats

import logging
logger = logging.getLogger(__name__)



def rotg(g1, g2, deg):
	"""
	Returns the rotated (g1, g2) by deg degrees.
	Be careful, there is this "extra" factor 2 here, given the meaning of g1 and g2.
	"""
	
	complexg = g1 + g2*1j
	rotg = complexg * np.exp(1j * 2.0 * np.pi*deg/180.0)
	return (rotg.real, rotg.imag)


def rmsd(x, y):
	"""
	Returns the RMSD between two numpy arrays
	(only beginners tend to inexactly call this the RMS... ;-)
	
	http://en.wikipedia.org/wiki/Root-mean-square_deviation
	
	This function also works as expected on masked arrays.	
	"""
	return np.sqrt(np.mean((x - y)**2.0))

def rmsd_delta(delta):
	"""
	Idem, but directly takes the estimation errors.
	Useful, e.g., to pass as reduce_C_function for hexbin plots.
	"""
	return np.sqrt(np.mean(np.asarray(delta)**2.0))


def linreg(x, y, prob=0.68):
	"""
	A linear regression y = m*x + c, with confidence intervals on m and c.
	
	As a safety measure, this function will refuse to work on masked arrays.
	Indeed scipy.stats.linregress() seems to silently disregard masks...
	... and as a safety measure, we compare against scipy.stats.linregress().
	
	"""
	
	if len(x) != len(y):
		raise RuntimeError("Your arrays x and y do not have the same size")
	
	if np.ma.is_masked(x) or np.ma.is_masked(y):
		raise RuntimeError("Do not give me masked arrays")
	
	n = len(x)
	xy = x * y
	xx = x * x
	
	b1 = (xy.mean() - x.mean() * y.mean()) / (xx.mean() - x.mean()**2)
	b0 = y.mean() - b1 * x.mean()
	
	#s2 = 1./n * sum([(y[i] - b0 - b1 * x[i])**2 for i in xrange(n)])
	s2 = np.sum((y - b0 - b1 * x)**2) / n
	
	alpha = 1.0 - prob
	c1 = scipy.stats.chi2.ppf(alpha/2.,n-2)
	c2 = scipy.stats.chi2.ppf(1-alpha/2.,n-2)
	#print 'the confidence interval of s2 is: ',[n*s2/c2,n*s2/c1]
	
	c = -1 * scipy.stats.t.ppf(alpha/2.,n-2)
	bb1 = c * (s2 / ((n-2) * (xx.mean() - (x.mean())**2)))**.5
	#print 'the confidence interval of b1 is: ',[b1-bb1,b1+bb1]
	
	bb0 = c * ((s2 / (n-2)) * (1 + (x.mean())**2 / (xx.mean() - (x.mean())**2)))**.5
	#print 'the confidence interval of b0 is: ',[b0-bb0,b0+bb0]
	
	ret = {"m":b1-1.0, "c":b0, "merr":bb1, "cerr":bb0}
	
	# A little test (for recent numpy, one would use np.isclose() for this !)
	slope, intercept, r_value, p_value, std_err = scipy.stats.linregress(x, y)
	if not abs(slope - b1) <= 1e-6 * abs(slope):
		raise RuntimeError("Slope error, %f, %f" % (slope, b1))
	if not abs(intercept - b0) <= 1e-6 * abs(intercept):
		raise RuntimeError("Intercept error, %f, %f" % (intercept, b0))
	
	return ret


# This function is unsufficiently documented, and the fact != a float is a joke (a bad one).
#
#def complex2geometrical(e1, e2, fact=1):
#	"""
#	Converts e1, e2 to a, b, theta
#	"""
#	e = np.sqrt(e1**2 + e2**2)
#	if fact != 1:
#		e = np.clip(e * fact, -0.9, 0.9)
#	a = np.sqrt(1. + e)
#	b = np.sqrt(1. - e )
#	theta = np.arctan2(e2, e1) / 2.
#	return a, b, theta





"""
This demo tests the network by asking for a plain regular regression (not inverse),
in 1D.

"""

import numpy as np
import tenbilac
import matplotlib.pyplot as plt


import logging
logging.basicConfig(level=logging.INFO)

#np.random.seed(0)

n = 200 # Size of training set
noise_scale = 0.05
ntest = 100 # Number of points to draw the regression line

def f(x):
	return np.sin(x)/(2. + x)

params = np.linspace(0.0, 10.0, n).reshape((1, n))
obs = f(params) + noise_scale*np.random.randn(n).reshape((1, n))

# Tenbilac only accept 3D input arrays
params = np.array([params])

# Normer object for the inputs and targets
obs_normer = tenbilac.data.Normer(obs)
params_normer = tenbilac.data.Normer(params)

# Do the normalisation
normobs = obs_normer(obs)
normparams = params_normer(params)

# Create some points where we want to see the value of the fitted obs, given the data.
testparams = np.linspace(-1.0, 11, ntest).reshape((1, ntest))
normtestparams = params_normer(testparams)

# Prepare a traindata object 
dat = tenbilac.data.Traindata(inputs=normparams, targets=normobs, valfrac=0.2)

# Generate the network
net = tenbilac.net.Net(1, [3])
training = tenbilac.train.Training(net, dat, errfctname="mse", regulweight=0.01, regulfctname="l2")

# Adding some (gaussian) noise to the weights and bias to initialise
training.net.addnoise(wscale=0.3, bscale=0.3)
# We train this normal (non-inverse) regression with params as inputs, and observations as output:
training.opt(algo="bfgs", mbsize=50, maxiter=500)

# Predicting the testparams
normtestpreds = net.run(normtestparams)
testpreds = obs_normer.denorm(normtestpreds)

print training

# We go back from a 3D array to 2D for plottting
params = params[0]

fig = plt.figure(figsize=(6, 4))

ax = fig.add_subplot(1, 1, 1)
ax.plot(params.T, obs.T, "b.", label="obs")
ax.plot(testparams.T, testpreds.T, "r-", label="fit")
ax.plot(testparams.T, f(testparams.T), "k--", lw=2, label="truth")
ax.set_xlabel(r"$x$", fontsize=18)
ax.set_ylabel(r"$y = f(x) + \mathrm{noise}$", fontsize=18)
ax.set_ylim([-0.5, 0.5])
ax.legend(loc='best')

plt.tight_layout()

tenbilac.plot.netviz(net)
plt.show()	




#import matplotlib
#matplotlib.use("AGG")

import momentsml
import momentsml.tools
import momentsmlgreat3
import astropy


import config
import numpy as np
import os


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)



subfields = []
tru_s1s = []
tru_s2s = []
pre_s1s = []
pre_s2s = []
pre_s1ws = [] # Not for the weights, but for the weighted averages!
pre_s2ws = []


psf_adamom_g1s = []
psf_adamom_g2s = []
psf_adamom_rho4s = []
psf_adamom_sigmas = []

tru_psf_g1s = []
tru_psf_g2s = []

for subfield in config.great3.subfields:
		
	predname = "pred_{}_{}".format(config.datasets["mimic-great3"], config.predcode) # Too messy to add everything here.
	predpath = config.great3.subpath(subfield, "pred", predname + ".pkl")
	
	if not os.path.exists(predpath):
		continue

	subfields.append(subfield)
	cat = momentsml.tools.io.readpickle(predpath)
	
	
	tru_s1s.append(cat["tru_s1"][0])
	tru_s2s.append(cat["tru_s2"][0])
		
	pre_s1s.append(np.ma.mean(cat["pre_s1"]))
	pre_s2s.append(np.ma.mean(cat["pre_s2"]))
	
	pre_s1ws.append(np.ma.mean(cat["pre_s1"] * cat["pre_s1w"]) / np.ma.mean(cat["pre_s1w"]))
	pre_s2ws.append(np.ma.mean(cat["pre_s2"] * cat["pre_s2w"]) / np.ma.mean(cat["pre_s2w"]))
	
	psf_adamom_g1s.append(np.mean(cat["psf_adamom_g1"])) # Yes, we take the mean here. We randomly distributed the measurements of 9 stars to the galaxies.
	psf_adamom_g2s.append(np.mean(cat["psf_adamom_g2"]))
	psf_adamom_rho4s.append(np.mean(cat["psf_adamom_rho4"]))
	psf_adamom_sigmas.append(np.mean(cat["psf_adamom_sigma"]))


	# We also need to read in the true PSF orientations (as measurements fail, sometimes)
	trustarpath = config.great3.trustarfilepath(subfield)
	trustar = momentsmlgreat3.io.readshear(trustarpath)
	assert len(trustar) == 4 # epoch_index psf_g1 psf_g2 subfield_index 
	assert int(subfield) == int(trustar[3])
	
	# Some of this PSF info was flagged, this is how they deal with it in the GREAT3 (SAD!)
	# We need to do the same, to reproduce GREAT3 metrics exactly.
	if trustar[1] < -9.9:
		trustar[1] = 0.0
	if trustar[2] < -9.9:
		trustar[2] = 0.0
	tru_psf_g1s.append(trustar[1])
	tru_psf_g2s.append(trustar[2])



# We make a catalog out of this

cat = astropy.table.Table([
		subfields, tru_s1s, tru_s2s, pre_s1s, pre_s2s, pre_s1ws, pre_s2ws,
		psf_adamom_g1s, psf_adamom_g2s, psf_adamom_rho4s, psf_adamom_sigmas,
		tru_psf_g1s, tru_psf_g2s
	], names=(
		'subfield', 'tru_s1', 'tru_s2', "pre_s1", "pre_s2", "pre_s1w", "pre_s2w",
		"psf_adamom_g1", "psf_adamom_g2", "psf_adamom_rho4", "psf_adamom_sigma",
		"tru_psf_g1", "tru_psf_g2"
		)
	)


#print momentsml.tools.table.info(cat)
catpath = config.great3.path("{}_summary_{}.pkl".format(config.datasets["mimic-great3"], config.predcode))
cat = momentsml.tools.io.writepickle(cat, catpath)



"""
This calls the GREAT3 evaluation function
"""


import momentsml
import momentsml.tools

import astropy

import config
import numpy as np
import os
import sys


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


outcatpath = config.great3.path("submission_{}.txt".format(config.predcode))

metricsdirpath = os.path.join(config.great3.g3publicdir, "metrics")
sys.path.append(metricsdirpath)
import evaluate

results = evaluate.q_constant(
	outcatpath, config.great3.experiment, config.great3.obstype,
	storage_dir=config.great3.path("eva_storage_dir_{}".format(config.predcode)),
	truth_dir=config.great3.truthdir,
	plot=config.great3.path("eva_plot_{}".format(config.predcode)),
	logger=logger,
	pretty_print=True)




"""
Regularisation functions
These work on the weights.

The typical call is fct(weights)
"""

import numpy as np

import logging
logger = logging.getLogger(__name__)


def l1(weights):
	"""
	The L1 norm is simply the sum of the absolute values. 
	"""
	return np.sum(np.abs(weights))

def l2(weights):
	"""
	The L2 norm is the sqrt of the sum of the squares. 
	"""
	return np.sqrt(np.sum(weights * weights))


"""
A measfct which computes statistics about the sky noise/level, using the edge of stamps

"""

import numpy as np
import copy
import astropy.table

import logging
logger = logging.getLogger(__name__)

import utils
from .. import tools


def measfct(cat, runon="img", prefix="", stampsize=None):
	"""
	
	:param runon: "img" or "psf" or ... -- decides on which image this should run.
		You might want to adjust the prefix accordingly.
	:param prefix: a prefix for the new column names. By default, no prefix is used.
	:param stampsize: if None, uses the image's stampsize.
	
	"""

	# Get the stampsize to be used:
	stampsize = cat.meta[runon].get_stampsize(stampsize)

	# Load the image:
	img = cat.meta[runon].load()
	
	# Prepare the catalog: THIS SHOULD BE REVIEWED ONCE WE ALL ADOPT ASTROPY 1.0 AND A RECENT NUMPY !
	cat = astropy.table.Table(copy.deepcopy(cat), masked=True) # Convert the table to a masked table
	# A bit stange: reading the doc I feel that this conversion is not needed.
	# But without it, it just doesn't result in a masked table once the masked columns are appended.

	cat.add_columns([
		astropy.table.Column(name=prefix+"skyflag", data=np.zeros(len(cat), dtype=int)), # We will always have a flag
		astropy.table.MaskedColumn(name=prefix+"skystd", dtype=float, length=len(cat)),
		astropy.table.MaskedColumn(name=prefix+"skymad", dtype=float, length=len(cat)),
		astropy.table.MaskedColumn(name=prefix+"skymean", dtype=float, length=len(cat)),
		astropy.table.MaskedColumn(name=prefix+"skymed", dtype=float, length=len(cat)),
		astropy.table.MaskedColumn(name=prefix+"skystampsum", dtype=float, length=len(cat)),
	])
	for col in ["skystd", "skymad", "skymean", "skymed", "skystampsum"]:
		cat[prefix+col].mask = [True] * len(cat)
		
	
	# And now the loop
	for gal in cat:
		
		# Some simplistic progress indication:
		if gal.index%5000 == 0:
			logger.info("%6.2f%% done (%i/%i) " % (100.0*float(gal.index)/float(len(cat)), gal.index, len(cat)))
		
		(x, y) = (gal[cat.meta[runon].xname], gal[cat.meta[runon].yname])
		(gps, flag) = tools.image.getstamp(x, y, img, stampsize)
		
		if flag != 0:
			logger.debug("Galaxy not fully within image:\n %s" % (str(gal)))
			gal[prefix+"skyflag"] = flag
		
		else:
			out = utils.skystats(gps)
			gal[prefix + "skystd"] = out["std"]
			gal[prefix + "skymad"] = out["mad"]
			gal[prefix + "skymean"] = out["mean"]
			gal[prefix + "skymed"] = out["med"]
			gal[prefix + "skystampsum"] = out["stampsum"]
			
	
	nfailed = np.sum(cat[prefix + "skyflag"] > 0)
	if nfailed != 0:
		logger.warning("The stamp extraction failed on %i out of %i sources (%.1f percent)" % (nfailed, len(cat), 100.0*float(nfailed)/float(len(cat))))
	
	return cat



"""
This is Tenbilac!
This file provides a class holder to define a ensemble (a committee) of Nets and a decorator class
that allows to train the committee on the same or different data. Committees are a stand-alone plugin.
Tenbilac Networks are not aware of this extension.
"""

import numpy as np
from multiprocessing import Process, Pipe, Queue
from itertools import izip  

import logging
logger = logging.getLogger(__name__)

from . import train
from . import utils


def spawn(f): 
	"""
	Helper function for `parmap` which prepares the worker `f` to be spawned.
	""" 
	def fun(pipe,x):  
		pipe.send(f(x))  
		pipe.close()  
	return fun  

def parmap(f, X, ncpu):  
	"""
	This is an alternative to multiprocessing.Pool to avoid the limitations of the package (pickling stuff...)
	
	.. note:: see http://stackoverflow.com/questions/3288595/multiprocessing-how-to-use-pool-map-on-a-function-defined-in-a-class
	
	.. note:: It is very possible that multiprocessing.Pool is fixed in python3 
	"""
	pipe=[Pipe() for x in X]  
	processes=[Process(target=spawn(f),args=(c,x)) for x,(p,c) in izip(X,pipe)]  
	numProcesses = len(processes)  
	processNum = 0  
	outputList = []  
	while processNum < numProcesses:  
		endProcessNum = min(processNum+ncpu, numProcesses)  
		for proc in processes[processNum:endProcessNum]:  
			proc.start()  
		# It is crucial to call recv() before join() to avoid deadlocks !
		for proc,c in pipe[processNum:endProcessNum]:  
			outputList.append(proc.recv())  
		for proc in processes[processNum:endProcessNum]:  
			proc.join()  

		processNum = endProcessNum  
	return outputList	

class Committee():
	"""
	Holder class for committee members
	"""
	
	def __init__(self, members, name=None):
		"""
		:param members: Net or MultNet etc instances
		:type members: tuple or list
		
		:param name: if None, will be set automatically
		:type name: string
		"""
		if name is None:
			self.name = 'Committee'
		else:
			self.name = name
		self.members = members
		
		logger.info("Built " + str(self))
		
		
	def __str__(self):
		"""
		A short string describing the network
		"""
		autotxt = "Committee class with {:d} members:\n".format(len(self.members))
		for ii, memi in enumerate(self.members):
			autotxt = "{autotxt}\t{ii}: {memi}\n".format(ii=ii+1, autotxt=autotxt, memi=memi)
		
		return "'{name}'\n{spacer}\n {autotxt}".format(name=self.name, spacer=(len(self.name)+2)*"=", autotxt=autotxt)
	
	def call(self, method, **kwargs):
		"""
		Method that runs a given `method` on all `trainings` instances.
		
		:param method: The name of the method
		:type method: string
		
		All other kwargs are passed to the method
		"""
		
		return np.ma.asarray([eval("memi.%s(**kwargs)" % (method)) for memi in self.members])


class CommTraining():
	"""
	This is an implicit decorator (??? wrapper -- or could you point to a tuto. I thought that decorators are functions returning functions)
	of the training class to handle the training of committees
	"""
	
	def __init__(self, committee, multiple_trainings=False, **kwargs):

		if multiple_trainings:
			# If multiple_trainings then it means that all args are a list of same len:
			lenkw = None
			for kw in kwargs:
				if len(committee.members) != len(kwargs[kw]): 
					raise ValueError("All kwargs must be of the same size as the committee! {name}: {lenk}/{lenc}".format(name=kw,\
						 lenk=len(kwargs[kw]), lenc=len(committee.members)))

		self.committee = committee
		rslt = []
		for ii, memi in enumerate(self.committee.members):
			
			if multiple_trainings:
				ukwargs = {}
				for kw in kwargs:
					ukwargs[kw] = kwargs[kw][ii]
			else:
				ukwargs = kwargs
			logger.info("Setting up Training committee {i}/{f}".format(i=(ii+1), f=len(self.committee.members)))
			output = train.Training(memi, **ukwargs)
			rslt.append(output)
			
		self.trainings = rslt
		
	def call(self, method, call_ncpu=1, attr=None, **kwargs):
		"""
		Method that runs a given `method` on all `trainings` instances.
		
		:param method: The name of the method
		:type method: string
		
		:param call_ncpu: number of cpu to use
		
		:param attr: calls a method on an attribute of all `trainings` instances. If `None` simply
			ignores this argument
			
		All other kwargs are passed to the method
		
		.. warning:: There is a known issue there. You should not run on multiple core except when training
			otherwise the callback function in training is not called.
			Could you elaborate:
			- why do I care that the callback function in training is not called, it I'm not training ?
			- is there any dangerous stuff left, for which the code would not raise an exception or something ?
			
		"""

		params = [[memi, method, attr, kwargs] for memi in self.trainings]
		
		if call_ncpu == 1:
			rslt = map(_worker, params)
		else:
			# We must use this, because of the limitations of multiprocessing.Pool
			rslt = parmap(_worker, params, call_ncpu)
			
		outputs = []
		
		# This is necessary for the multiple cpu case 
		for ii, (rs, out) in enumerate(rslt):
			self.trainings[ii] = rs 
			self.committee.members[ii] = rs.net
			outputs.append(out)
			
		outputs = np.ma.asarray(out)

		return outputs 
	
	def save(self, filepath, keepdata=False):
		"""
		Saves the training committee into a pkl file
		As the training data is so massive, by default we do not save it!
		Note that this might be done at each iteration!
		
		..note:: This is a similar method as in `train.py`
		"""

		if keepdata is True:
			logger.info("Writing training committee to disk and keeping the data...")
			utils.writepickle(self, filepath)	
		else:
			tmptraindata = [dat for dat in self.trainings]
			self.call('set_dat', dat=None)
				
			utils.writepickle(self, filepath)	
			
			for ii in range(len(self.trainings)):
				self.trainings[ii].dat = tmptraindata[ii]
						
				

def _worker(params):
	training, method, attr, kwargs = params

	if attr is None:
		cmd_line = "training.%s(**kwargs)" % (method)
	else:
		cmd_line = "training.%s.%s(**kwargs)" % (attr, method)
	output = eval(cmd_line)

	return training, output







import numpy as np
import tenbilac


import logging
logging.basicConfig(level=logging.INFO)



layer = tenbilac.layer.Layer(ni=2, nn=3, actfct=tenbilac.act.iden)

#layer.addnoise()

# 3D case (i.e., with several realizations)

layer.weights[0, 0] = 1
layer.weights[0, 1] = 2


layer.weights[2, 0] = 3
layer.weights[2, 1] = 0
layer.biases[1] = 42


print layer.report()

# 2 features, 2 galaxies, 4 realizations:
# galaxy_1 is (1, -1), galaxy_2 is (4, -4)
# order of indices: realization, feature, galaxy
input = np.array([
[[1, 4], [-1, -4]],
[[1.1, 4], [-1.3, -4]],
[[1, 4], [-1, -4]],
[[1, 4], [-1, -4]]
])
# Checking the galaxy 1 and 2:
print input[0, 0, 0], input[0, 1, 0]
print input[0, 0, 1], input[0, 1, 1]


print input.shape
#print input.ndim


output = layer.run(input)


print "ouput = ", output
print output.shape


# 2 D case:
"""
layer.weights[1, 1] = 1
layer.biases[2] = 10
print layer.report()

# 2 features, 4 galaxies. Each galaxy is (1, 2)
input = np.array([[1, 1, 1, 1], [2, 2, 2, 2]])
#input = np.array([1, 2])

print "input = ", input
output = layer.run(input)
print "ouput = ", output
"""



import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os

import plot_3_valw

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)



for subfield in config.great3.subfields:

	logger.info("Working on subfield {}".format(subfield))

	catpath = config.great3.subpath(subfield, "simmeas", config.datasets["mimic-great3"], "groupmeascat.pkl")
	
	if not os.path.exists(catpath):
		continue
	
	cat = momentsml.tools.io.readpickle(catpath)
	#print cat["tru_rad", "tru_read_noise", "psf_adamom_g1", "tru_s1", "tru_g1"]
	#print momentsml.tools.table.info(cat)
	#exit()
	
	sheartraindir = config.great3.subpath(subfield, "ml", config.datasets["train-shear"])
	cat = momentsml.learn.tenbilacrun.predict(cat, config.shearconflist , sheartraindir)
	
	if len(config.weightconflist) > 0:
		
		weighttraindir = config.great3.subpath(subfield, "ml", config.datasets["train-weight"])
		cat = momentsml.learn.tenbilacrun.predict(cat, config.weightconflist , weighttraindir)
		
	
	predname = "pred_{}_{}".format(config.datasets["mimic-great3"], config.predcode) # Too messy to add everything here.
	predpath = config.great3.subpath(subfield, "pred", predname + ".pkl")
	momentsml.tools.io.writepickle(cat, predpath)




import config
import measfcts
import os
import numpy as np
import matplotlib.ticker as ticker
import matplotlib.colors
import matplotlib

import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt
plt.rc('text', usetex=True)

import logging
logger = logging.getLogger(__name__)


#momentsml.plot.figures.set_fancy(14)
from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)

rasterized=True

valcat = os.path.join(config.simmeasdir, config.datasets["vp"], "groupmeascat.pkl")
cat = momentsml.tools.io.readpickle(valcat)

#print momentsml.tools.table.info(cat)

momentsml.tools.table.addstats(cat, "snr")

cat["adamom_goodfrac"] = 1.0 - cat["adamom_failfrac"]


tru_sb = Feature("tru_sb", 0, 16, nicename=r"Surface brightness $S$ [pix$^{-2}$]")
tru_rad = Feature("tru_rad", 1.5, 8.5, nicename=r"Half-light radius $R$ [pix]")

snr_mean = Feature("snr_mean", nicename="$\\langle \mathrm{S}/\mathrm{N}\\rangle$")

successfrac = Feature("adamom_goodfrac", nicename="Measurement success fraction")

for comp in ["1"]:

	
	fig = plt.figure(figsize=(3.5, 6))

	ax = fig.add_subplot(2, 1, 1)
	cb = momentsml.plot.scatter.scatter(ax, cat, tru_sb, tru_rad, featc=snr_mean, cmap="plasma_r", norm=matplotlib.colors.PowerNorm(gamma=1./2.), rasterized=rasterized)
	
	cb.set_ticks([2, 5, 10, 20, 30, 40, 50, 60])
	
	ax = fig.add_subplot(2, 1, 2)
	momentsml.plot.scatter.scatter(ax, cat, tru_sb,  tru_rad, featc=successfrac, cmap="plasma_r", rasterized=rasterized)
	ax.axhline()
	

	
	plt.tight_layout()


	momentsml.plot.figures.savefig(os.path.join(config.valdir, config.datasets["vs"] + "_snr"), fig, fancy=True, pdf_transparence=True)
	plt.show()




import numpy as np

import matplotlib.pyplot as plt

def contracted_rayleigh(sigma, max_val, p):
	"""
	A distribution with finite support.
	"""
	tmp = np.random.rayleigh(sigma)
	return (tmp / np.power(1 + np.power(tmp / max_val, p), 1.0 / p))

def trun_rayleigh(sigma, max_val):
	"""
	A truncated Rayleigh distribution
	"""
	assert max_val > sigma
	tmp = max_val + 1.0
	while tmp > max_val:
		tmp = np.random.rayleigh(sigma)
	return tmp


n = 500000
sigma = 0.2
bins = 100

rs = [np.random.rayleigh(sigma) for i in range(n)]
crs = [contracted_rayleigh(sigma, 0.7, 4) for i in range(n)]
trs = [trun_rayleigh(sigma, 0.7) for i in range(n)]


plt.hist(rs, label="Rayleigh", range=(0, 1.0), alpha=0.5, bins=bins)
plt.hist(crs, label="Contracted Rayleigh", range=(0, 1.0), alpha=0.5, bins=bins)
plt.hist(trs, label="Truncated Rayleigh", range=(0, 1.0), alpha=0.5, bins=bins)

plt.legend()
plt.show()



"""
Plots the figure of the paper 
"""

import config
import measfcts
import os
import numpy as np
import matplotlib.ticker as ticker

import momentsml.plot
from momentsml.tools.feature import Feature
import matplotlib.pyplot as plt

import logging
logger = logging.getLogger(__name__)



#momentsml.plot.figures.set_fancy(14)
from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)



mode = "vo"



if mode == "vs":
	valname = config.valname
elif mode == "vo":
	valname = config.wvalname

valcatpath = os.path.join(config.valdir, valname + ".pkl")
cat = momentsml.tools.io.readpickle(valcatpath)


"""
newmask = cat["tru_rad"] < 1.0
logger.info("Fraction of gals with rad < 1")
logger.info(float(np.sum(newmask)) / np.size(newmask) )
origmask = cat["pre_s1"].mask
combimask = np.logical_or(newmask, origmask)
cat["pre_s1"].mask = combimask
cat["pre_s2"].mask = combimask
cat["pre_s1w"].mask = combimask
cat["pre_s2w"].mask = combimask
"""


#print momentsml.tools.table.info(cat)

for comp in ["1","2"]:

	# This is an interesting experiement. it reminds us how HUGE selection biases are...
	#snrcutmask = cat["snr"] < 10.0
	#cat["pre_s1w"][snrcutmask] = 0.0
	#cat["pre_s2w"][snrcutmask] = 0.0
	

	# If no weights are in the catalog (or not yet), we add ones
	if not "pre_s{}w".format(comp) in cat.colnames:
		
		# First putting all weights to 1.0:
		cat["pre_s{}w".format(comp)] = np.ones(cat["adamom_g1"].shape)
		logger.info("Setting weights to one")	
	
	cat["pre_s{}w_norm".format(comp)] = cat["pre_s{}w".format(comp)] / np.max(cat["pre_s{}w".format(comp)])

	momentsml.tools.table.addrmsd(cat, "pre_s{}".format(comp), "tru_s{}".format(comp))
	momentsml.tools.table.addstats(cat, "pre_s{}".format(comp), "pre_s{}w".format(comp))
	cat["pre_s{}_wbias".format(comp)] = cat["pre_s{}_wmean".format(comp)] - cat["tru_s{}".format(comp)]


resr = 0.01
symthres = 0.002
wplotrea = -10
snr = Feature("snr", nicename="S/N", rea=wplotrea)
#tru_rad = Feature("tru_rad", nicename=r"$R$ [pix]", rea=wplotrea)

tru_mag = Feature("tru_mag", 20, 26, nicename="Magnitude", rea=wplotrea)

tru_rad = Feature("tru_rad", 1.5, 8.5, nicename=r"Half-light radius $R$ [pix]", rea=wplotrea)

adamom_flux = Feature("adamom_flux", 0, 1000, nicename="adamom\_flux", rea=wplotrea)
adamom_sigma = Feature("adamom_sigma", 0, 8, nicename="adamom\_sigma", rea=wplotrea)
adamom_g2 = Feature("adamom_g2", nicename="adamom\_g2", rea=wplotrea)
adamom_rho4 = Feature("adamom_rho4", nicename="adamom\_rho4", rea=wplotrea)


tru_s1 = Feature("tru_s1", nicename=r"$g_1^{\mathrm{true}}$")
tru_s2 = Feature("tru_s2", nicename=r"$g_2^{\mathrm{true}}$")

pre_s1_bias = Feature("pre_s1_bias", -resr, resr, nicename=r"$\langle \hat{g}_{1} \rangle - g_{1}^{\mathrm{true}} $")
pre_s2_bias = Feature("pre_s2_bias", -resr, resr, nicename=r"$\langle \hat{g}_{2} \rangle - g_{2}^{\mathrm{true}} $")
pre_s1_wbias = Feature("pre_s1_wbias", -resr, resr, nicename=r"$\left(\sum\hat{g}_1 w_1 / \sum w_1 \right) - g_{1}^{\mathrm{true}}$")
pre_s2_wbias = Feature("pre_s2_wbias", -resr, resr, nicename=r"$\left(\sum\hat{g}_2 w_2 / \sum w_2 \right) - g_{2}^{\mathrm{true}}$")



#pre_s1_bias = Feature("pre_s1_bias", -resr, resr, nicename=r"Bias on $\hat{g}_{1}$")
#pre_s2_bias = Feature("pre_s2_bias", -resr, resr, nicename=r"Bias on $\hat{g}_{2}$")
#pre_s1_wbias = Feature("pre_s1_wbias", -resr, resr)
#pre_s2_wbias = Feature("pre_s2_wbias", -resr, resr)


def addmetrics(ax, xfeat, yfeat):
	metrics = momentsml.tools.metrics.metrics(cat, xfeat, yfeat, pre_is_res=True)
	line1 = r"$\mathrm{RMSD} = %.5f $" % (metrics["rmsd"])
	#line2 = r"$m: %.1f +/- %.1f; c: %.1f +/- %.1f$" % (metrics["m"]*1000.0, metrics["merr"]*1000.0, metrics["c"]*1000.0, metrics["cerr"]*1000.0)
	#line2 = r"$m=%.1f \pm %.1f; c=%.1f \pm %.1f \, [10^{-3}]$" % (metrics["m"]*1000.0, metrics["merr"]*1000.0, metrics["c"]*1000.0, metrics["cerr"]*1000.0)
	line2 = r"$10^3 \mu=%.1f \pm %.1f $" % (metrics["m"]*1000.0, metrics["merr"]*1000.0)
	line3 = r"$10^3 c=%.1f \pm %.1f $" % (metrics["c"]*1000.0, metrics["cerr"]*1000.0)
	
	ax.annotate(line1, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -9), textcoords='offset points', ha='left', va='top', fontsize=12)
	ax.annotate(line2, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -22), textcoords='offset points', ha='left', va='top', fontsize=12)
	ax.annotate(line3, xy=(0.0, 1.0), xycoords='axes fraction', xytext=(8, -36), textcoords='offset points', ha='left', va='top', fontsize=12)


fig = plt.figure(figsize=(11.5, 7))
plt.subplots_adjust(
	left  = 0.08,  # the left side of the subplots of the figure
	right = 0.90,    # the right side of the subplots of the figure
	bottom = 0.1,   # the bottom of the subplots of the figure
	top = 0.90,      # the top of the subplots of the figure
	wspace = 0.30,   # the amount of width reserved for blank space between subplots,
	                # expressed as a fraction of the average axis width
	hspace = 0.25,   # the amount of height reserved for white space between subplots,
					# expressed as a fraction of the average axis heightbottom=0.1, right=0.8, top=0.9)
	)

idlinekwargs = {"color":"black", "ls":"-"}

#==================================================================================================

ax = fig.add_subplot(2, 3, 1)
momentsml.plot.scatter.scatter(ax, cat, tru_s1, pre_s1_bias, showidline=True, idlinekwargs=idlinekwargs, yisres=True)
#ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
addmetrics(ax, tru_s1, pre_s1_bias)
ax.set_title("Without weights")
ax.title.set_position([.5, 1.1])
#ax.set_xlabel("")
#ax.set_xticklabels([])


ax = fig.add_subplot(2, 3, 2)
momentsml.plot.scatter.scatter(ax, cat, tru_s1, pre_s1_wbias, showidline=True, idlinekwargs=idlinekwargs, yisres=True)
#ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
ax.set_title("With weights")
ax.title.set_position([.5, 1.1])
addmetrics(ax, tru_s1, pre_s1_wbias)
#ax.set_xlabel("")
#ax.set_xticklabels([])
#ax.set_ylabel("")
ax.set_yticklabels([])


ax = fig.add_subplot(2, 3, 3)
pos1 = ax.get_position() # get the original position 
pos2 = [pos1.x0 + 0.05, pos1.y0,  pos1.width, pos1.height] 
ax.set_position(pos2)
momentsml.plot.scatter.scatter(ax, cat, adamom_flux, adamom_sigma, featc=Feature("pre_s1w_norm", 0, 1, rea=wplotrea, nicename=r"Weight $w_1$"), cmap="plasma_r", rasterized = True)
#momentsml.plot.scatter.scatter(ax, cat, tru_sb, tru_rad, featc=Feature("pre_s1w_norm", 0, 1, rea=wplotrea, nicename=r"Weight $w_1$"), cmap="plasma_r", rasterized = True)


#==================================================================================================

ax = fig.add_subplot(2, 3, 4)
momentsml.plot.scatter.scatter(ax, cat, tru_s2, pre_s2_bias, showidline=True, idlinekwargs=idlinekwargs, yisres=True)
#ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
addmetrics(ax, tru_s2, pre_s2_bias)


ax = fig.add_subplot(2, 3, 5)
momentsml.plot.scatter.scatter(ax, cat, tru_s2, pre_s2_wbias, showidline=True, idlinekwargs=idlinekwargs, yisres=True)
#ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
addmetrics(ax, tru_s2, pre_s2_wbias)
#ax.set_ylabel("")
ax.set_yticklabels([])


ax = fig.add_subplot(2, 3, 6)
pos1 = ax.get_position() # get the original position 
pos2 = [pos1.x0 + 0.05, pos1.y0,  pos1.width, pos1.height] 
ax.set_position(pos2)
#momentsml.plot.scatter.scatter(ax, cat, adamom_flux, adamom_sigma, featc=Feature("pre_s2w_norm", 0, 1, rea=wplotrea, nicename=r"Weight $w_2$"), cmap="plasma_r", rasterized = True)
momentsml.plot.scatter.scatter(ax, cat, tru_mag, tru_rad, featc=Feature("pre_s2w_norm", 0, 1, rea=wplotrea, nicename=r"Weight $w_2$"), cmap="plasma_r", rasterized = True)


#==================================================================================================

#plt.tight_layout()

momentsml.plot.figures.savefig(os.path.join(config.valdir, valname + "_wstudy"), fig, fancy=True, pdf_transparence=True)
plt.show()





import matplotlib
matplotlib.use("AGG")

import argparse

import momentsml.tools
import momentsml.learn
import momentsml
import tenbilac

import config
import numpy as np
import os
import glob
import shutil

import plot_3_valw


import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)


parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
parser.add_argument('-s', '--startfrom', type=int, help='Subfield from which to start the training')
args = parser.parse_args()


for subfield in config.great3.subfields:
	
	
	logger.info("Working on subfield {}".format(subfield))
	
	# Getting the path to the correct directories
	measdir = config.great3.subpath(subfield, "simmeas", config.datasets["train-weight"])
	traindir = config.great3.subpath(subfield, "ml", config.datasets["train-weight"])
	
	# And to the catalogue
	traincatpath = os.path.join(measdir, "groupmeascat_predforw.pkl")
	cat = momentsml.tools.io.readpickle(traincatpath)
		
	
	# Now we might want to copy an existing training to start from there.
	if args.startfrom is not None:
		logger.info("Copying existing training from {}...".format(args.startfrom))
	
		dirnames = momentsml.learn.tenbilacrun.confnames(config.weightconflist)
		for (dirname, conf) in zip(dirnames, config.weightconflist):
			newtraindirpath = os.path.join(traindir, dirname)
			if os.path.isdir(newtraindirpath):
				raise RuntimeError("A training already exists, but you want me to start from another one")
			
			#pretrainminipaths = sorted(glob.glob(config.great3.subpath(args.startfrom, "ml", config.datasets["train-weight"], dirname, "mini_*")))
			#if len(pretrainminipaths) == 0:
			#	raise RuntimeError("Could not find pre-training!")
			#pretrainpath = pretrainminipaths[-1] # Taking the last one
			
			# Let's not take mini, but the full one
			pretrainpath = config.great3.subpath(args.startfrom, "ml", config.datasets["train-weight"], dirname)

			logger.info("Copying {} into {}...".format(pretrainpath, newtraindirpath))
			shutil.copytree(pretrainpath, newtraindirpath)

			print "##############################   Is it ok to keep the normer ?"
	
	# Running the training	
	dirnames = momentsml.learn.tenbilacrun.train(cat, config.weightconflist, traindir)
	#dirnames = ["ada4s1w_sum33wmassshort", "ada4s2w_sum33wmassshort"]
	
	# Summary plot and self-predictions
	for dirname in dirnames:
		tenbilacdirpath = os.path.join(traindir, dirname)
		ten = tenbilac.com.Tenbilac(tenbilacdirpath)
		ten._readmembers()
		tenbilac.plot.summaryerrevo(ten.committee, filepath=os.path.join(traindir, "{}_summary.png".format(dirname)))


	cat = momentsml.tools.io.readpickle(traincatpath) # To get the full catalog with all cases	
	
	for (dirname, conf) in zip(dirnames, config.weightconflist):

		predcat = momentsml.learn.tenbilacrun.predict(cat, [conf], traindir)
		predcatpath = os.path.join(traindir, "{}_selfpred.pkl".format(dirname))
		momentsml.tools.io.writepickle(predcat, predcatpath)
		figpredcatpath = os.path.join(traindir, "{}_selfpred.png".format(dirname))
		
		if "s2" in conf[0] or "g2" in conf[0]:
			component = 2
		else:
			component = 1
		plot_3_valw.plot(predcat, component, filepath=figpredcatpath)
			



"""
This is a standalone wrapper for Tenbilac, featuring:

  - a simpler configuration, making it easy to explore, run and use different trainings
  - a no-brainer way to run several trainings one after the other (no multiprocessing, as tenbilac does this internally).


"""
import datetime
import numpy as np
import astropy
import copy

import os
from ConfigParser import SafeConfigParser
from .. import tools

import tenbilac

import logging
logger = logging.getLogger(__name__)




def train(catalog, conflist, workbasedir):
    """
    Top-level function to train Tenbilacs with data from a MomentsML catalog.
    Does not modify or return the input catalog, it really just trains.
    
    :param conflist: A list of tuples ("ada5g1.cfg", "sum55.cfg") of filepaths to the configuration files for the data selection and the machine learning.
        Tuples in this list are processed one after the other.
    :param workbasedir: A directory in which the trainings can be organized.
        I will create a subdirectory in there reflecting the configuration names.
    
    It returns a list of names of the Tenbilac workdirs (as long as conflist).
    This can be useful for example if you want to save plots, catalogs or (self-)predictions in there automatically after the training.    
    """
    starttime = datetime.datetime.now()
    logger.info("Starting the training of {} MLs...".format(len(conflist)))
    
    trainworkdirs = []
    for (dataconfpath, toolconfpath) in conflist:
        
        # We read in the configurations
        dataconfig = readconfig(dataconfpath) # The data config (what to train usign which features)
        toolconfig = readconfig(toolconfpath) # The Tenbilac config
        confname = dataconfig.get("setup", "name") + "_" + toolconfig.get("setup", "name") # Will be passed to Tenbilac
        trainworkdir = os.path.join(workbasedir, confname) # We will pass this to Tenbilac
        trainworkdirs.append(confname)
        
        # We get the config in form of python lists:
        inputlabels = list(eval(dataconfig.get("data", "inputlabels")))
        auxinputlabels = list(eval(dataconfig.get("data", "auxinputlabels")))
        targetlabels = list(eval(dataconfig.get("data", "targetlabels")))
        
        # Preparing the inputs
        inputsdata = get3Ddata(catalog, inputlabels)
        if len(auxinputlabels) != 0:
            logger.info("We have auxinputs")
            auxinputsdata = get3Ddata(catalog, auxinputlabels)
        else:
            logger.info("No auxinputs")
            auxinputsdata = None
        
        # Preparing the targets
        for colname in targetlabels:
            if not np.all(np.logical_not(np.ma.getmaskarray(catalog[colname]))): # No element should be masked.
                raise RuntimeError("Targets should not be masked, but '{}' is!".format(colname))
        targetsdata = np.column_stack([np.array(catalog[colname]) for colname in targetlabels]).transpose()
        assert inputsdata.shape[2] == targetsdata.shape[1] # Number of cases should match
        if auxinputsdata is not None:
            assert auxinputsdata.shape[2] == targetsdata.shape[1] #  Number of cases must match!
                # note that these tests are repeated withing tenbilac, of course.            

        
        # And calling Tenbilac
        tblconfiglist = [("setup", "workdir", trainworkdir), ("setup", "name", confname)]
        ten = tenbilac.com.Tenbilac(toolconfpath, tblconfiglist)
        
        ten.train(inputsdata, targetsdata, inputlabels, targetlabels, auxinputs=auxinputsdata)
        
    
        
    endtime = datetime.datetime.now()
    logger.info("Done, the total time for training the %i MLs was %s" % (len(conflist), str(endtime - starttime)))
    
    return trainworkdirs


def confnames(conflist):
    """
    Returns a list of names for the conflist items, which can be used as directories.
    This is exactly what train also returns.
    """
    output = []
    for (dataconfpath, toolconfpath) in conflist:
        # We read in the configurations
        dataconfig = readconfig(dataconfpath) # The data config (what to train usign which features)
        toolconfig = readconfig(toolconfpath) # The Tenbilac config
        confname = dataconfig.get("setup", "name") + "_" + toolconfig.get("setup", "name")
        output.append(confname)
    return output

def plot_fancy(conflist, workbasedir=None):

    for (dataconfpath, toolconfpath) in conflist:
        
        # We read in the configurations, as for the training, to get the Tenbilac workdir to use.
        dataconfig = readconfig(dataconfpath)
        if os.path.isfile(toolconfpath):
            toolconfig = readconfig(toolconfpath)
            if workbasedir is None:
                raise RuntimeError("As config files (and not dirs) are given to predict, it also needs a workbasedir!")
            trainworkdir = os.path.join(workbasedir, dataconfig.get("setup", "name") + "_" + toolconfig.get("setup", "name"))
        elif os.path.isdir(toolconfpath):
            # Then we don't have to read this now. Tenbilac will take care of it.
            trainworkdir = toolconfpath
        else:
            raise RuntimeError("Tenbilac config '{}' does not exist.".format(toolconfpath))
        logger.info("Preparing for predictions usign the Tenbilac workdir '{}'...".format(trainworkdir))
        
        # And calling Tenbilac
        tblconfiglist = [("setup", "workdir", trainworkdir)]
        ten = tenbilac.com.Tenbilac(toolconfpath, tblconfiglist)
        
        ten.plot_fancy()

def predict(catalog, conflist, workbasedir=None):
    """
    Top level-function to make predictions. The only "tricky" thing here is that it returns a single catalog with all the new columns.
    This catalog is built progressively: the second machine in your configlist could potentially use predictions from the first one as inputs.
    
    The input catalog is not modified.
    
    If any feature values of your catalog are masked, the corresponding rows in the output catalog will not be predicted,
    and the prediction columns will get masked accordingly.
    
    Same arguments as for train, but conflist can *alternatively* consist of tuples ("ada5g1.cfg", "/path/to/tenbilac-workdir/to/use").
        If the tenbilac configuration is given as a directory, then this directory is used to make predictions, intead of the
        combination of workbasedir and the configuration names as done during the training.
        The argument workbasedir is ignored, in this case. This feature makes it much simpler and safer to write standalone scripts 
        that apply a bunch of trained Tenbilacs to data.
    
    """
    starttime = datetime.datetime.now()
    logger.info("Starting predictions from {} MLs...".format(len(conflist)))
    
    outcat = astropy.table.Table(copy.deepcopy(catalog), masked=True)
    
    for (dataconfpath, toolconfpath) in conflist:
        
        # We read in the configurations, as for the training, to get the Tenbilac workdir to use.
        dataconfig = readconfig(dataconfpath)
        if os.path.isfile(toolconfpath):
            toolconfig = readconfig(toolconfpath)
            if workbasedir is None:
                raise RuntimeError("As config files (and not dirs) are given to predict, it also needs a workbasedir!")
            trainworkdir = os.path.join(workbasedir, dataconfig.get("setup", "name") + "_" + toolconfig.get("setup", "name"))
        elif os.path.isdir(toolconfpath):
            # Then we don't have to read this now. Tenbilac will take care of it.
            trainworkdir = toolconfpath
        else:
            raise RuntimeError("Tenbilac config '{}' does not exist.".format(toolconfpath))
        logger.info("Preparing for predictions usign the Tenbilac workdir '{}'...".format(trainworkdir))
        
        # We get the required data config in form of python lists:
        inputlabels = list(eval(dataconfig.get("data", "inputlabels")))
        predlabels = list(eval(dataconfig.get("data", "predlabels")))
        
        # Check that the predictions do not yet exist
        for predlabel in predlabels:
            if predlabel in catalog.colnames:
                raise RuntimeError("The predlabel '%s' already exists in the catalog!" % predlabel)
        
        # Preparing the inputs
        inputsdata = get3Ddata(outcat, inputlabels)
        
        # And calling Tenbilac
        tblconfiglist = [("setup", "workdir", trainworkdir)]
        ten = tenbilac.com.Tenbilac(toolconfpath, tblconfiglist)
        
        preddata = ten.predict(inputsdata)
        
        # We add this to the outcat.
        # An explicit loop, to highlight that we care very much about the order (to get targetlabels right)
            
        for (i, predlabel) in enumerate(predlabels):    
            logger.info("Adding predictions '{}' to catalog...".format(predlabel))
            data = preddata[:,i,:].transpose()
            assert data.ndim == 2 # Indeed this is now always 2D.
            if data.shape[1] == 1: # If we have only one realization, just make it a 1D numpy array.
                data = data.reshape((data.size))
                assert data.ndim == 1
                        
            newcol = astropy.table.MaskedColumn(data=data, name=predlabel)
            outcat.add_column(newcol)
            
    endtime = datetime.datetime.now()
    logger.info("Done, the total time for the predictions was {}".format(str(endtime - starttime)))

    return outcat


def readconfig(configpath):
    """
    Reads in a config file
    """
    config = SafeConfigParser(allow_no_value=True)
    
    if not os.path.exists(configpath):
        raise RuntimeError("Config file '{}' does not exist!".format(configpath))
    logger.debug("Reading config from '{}'...".format(configpath))
    config.read(configpath)
    
    name = config.get("setup", "name") 
    if name is None or len(name.strip()) == 0: # if the ":" is missing as well, confirparser reads None
        # Then we use the filename
        config.set("setup", "name", os.path.splitext(os.path.basename(configpath))[0])
    logger.info("Read config '{}' from file '{}'.".format(config.get("setup", "name"), configpath))    
    
    return config
    

def get3Ddata(catalog, colnames):
    """
    Function to build a 3D numpy array (typically for Tenbilac input) from some columns of an astropy catalog.
    The point is to ensure that all columns get the same shape.
    
    The 3D output array has shape (realization, feature, case).
    """
    
    if len(colnames) == 0:
        raise RuntimeError("No colnames to get data from!")
    
    # Check for exotic catalogs (do they even exist ?)
    for colname in colnames:
        if not catalog[colname].ndim in [1, 2]:
            raise RuntimeError("Can only work with 1D or 2D columns")
    
    # Let's check the depths of the 2D colums to see what size we need.
    nreas = list(set([catalog[colname].shape[1] for colname in colnames if catalog[colname].ndim == 2]))
    #logger.info("We have the following nreas different from one in there: {}".format(nreas))
    if len(nreas) > 1:
        raise RuntimeError("The columns have incompatible depths!")

    if len(nreas) == 0:
        nrea = 1
        logger.info("For each column, only one realization is available.")
        
    else:
        nrea = nreas[0]
        logger.info("Extracting data from {0} realizations...".format(nrea))
        nrea = nreas[0]
    
    if "ngroup" in catalog.meta:
        if nrea != catalog.meta["ngroup"]:
            raise RuntimeError("Something very fishy: depth is not ngroup!")

    # And now we get the data:
    
    readycols = []
    for colname in colnames:
                
        col = np.ma.array(catalog[colname])
                
        if col.ndim == 2:
            pass
            
        elif col.ndim == 1:
            # This column has only one realization, and we have to "duplicate" it nrea times...
            col = np.tile(col, (nrea, 1)).transpose()
                    
        else:
            raise RuntimeError("Weird column dimension")
                                
        readycols.append(col)
        
    outarray = np.rollaxis(np.ma.array(readycols), 2)
    
    assert outarray.ndim == 3
    assert outarray.shape[1] == len(colnames)

    return outarray



import momentsml
import os
import config
import numpy as np

from momentsml.tools.feature import Feature

import matplotlib
import matplotlib.pyplot as plt



#cat = momentsml.tools.io.readpickle(os.path.join(config.simmeasdir, "si-1-gems-20rea", "groupmeascat.pkl"))
cat = momentsml.tools.io.readpickle(os.path.join(config.simmeasdir, "si-1-uni-100rea", "groupmeascat.pkl"))


momentsml.tools.table.addstats(cat, "snr")

#print momentsml.tools.table.info(cat)
#exit()

fig = plt.figure(figsize=(20, 13))

snr_mean = Feature("snr_mean", 0, 20, nicename="$\\langle \mathrm{S}/\mathrm{N}\\rangle$")

snr = Feature("snr", 0, 50, nicename="Measured S/N")
tru_rad = Feature("tru_rad", nicename="True half-light radius [pixel]")

tru_mag = Feature("tru_mag", 20, 26, nicename="Magnitude")
tru_sersicn = Feature("tru_sersicn", 0, 7, nicename="True Sersic index")

skystd = Feature("skystd")
skymed = Feature("skymed")
adamom_flux = Feature("adamom_flux")
adamom_logflux = Feature("adamom_logflux")
adamom_sigma = Feature("adamom_sigma")
adamom_g1 = Feature("adamom_g1")
adamom_g2 = Feature("adamom_g2")
adamom_g = Feature("adamom_g")
adamom_rho4 = Feature("adamom_rho4")


ax = fig.add_subplot(2, 3, 1)
momentsml.plot.scatter.scatter(ax, cat, tru_mag,  tru_rad, sidehists=True, sidehistkwargs={"bins":20})

ax = fig.add_subplot(2, 3, 2)
momentsml.plot.scatter.scatter(ax, cat, tru_mag,  snr_mean, tru_rad)

ax = fig.add_subplot(2, 3, 3)
momentsml.plot.scatter.scatter(ax, cat, tru_rad,  snr_mean, tru_mag)

ax = fig.add_subplot(2, 3, 4)
momentsml.plot.scatter.scatter(ax, cat, tru_rad, tru_mag, snr_mean)

plt.tight_layout()
plt.show()
plt.close(fig) # Helps releasing memory when calling in large loops.





import matplotlib
matplotlib.use("AGG")

import momentsml.tools
import momentsml.learn
import momentsml

import config
import numpy as np
import os

import plot_2_val

import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)






for subfield in config.great3.subfields:

	logger.info("Working on subfield {}".format(subfield))

	for conf in config.shearconflist: # We go through them one by one here.

		catpath = config.great3.subpath(subfield, "simmeas", config.datasets["valid-shear"], "groupmeascat.pkl")
		cat = momentsml.tools.io.readpickle(catpath)
		traindir = config.great3.subpath(subfield, "ml", config.datasets["train-shear"])

		confname = momentsml.learn.tenbilacrun.confnames([conf])[0]
		predcat = momentsml.learn.tenbilacrun.predict(cat, [conf], traindir)

		valname = "pred_{}_with_{}_{}".format(config.datasets["valid-shear"], config.datasets["train-shear"], confname)	

		predcatpath = config.great3.subpath(subfield, "val", valname + ".pkl")
		figpredcatpath = config.great3.subpath(subfield, "val", valname + ".png")
		
		figgoodpredcatpath = config.great3.subpath(subfield, "val", valname + "_good.png")
		predcat2 = predcat.copy()
			
		momentsml.tools.io.writepickle(predcat, predcatpath)
	
		# And make the plot	
		if "s2" in conf[0] or "g2" in conf[0]:
			component = 2
		else:
			component = 1
		plot_2_val.plot(predcat, component, mode="s", filepath=figpredcatpath)
		
		s = momentsml.tools.table.Selector("good", [
			("max", "pre_maskedfrac", 0.1),
		])
		plot_2_val.plot(predcat2, component, mode="s", select=s, filepath=figgoodpredcatpath)
	
	"""
	conflist = [
		#("mlconfig/ada4s1.cfg", config.great3.path("ml", "%03i" % subfield, trainspname, "ada4s1_sum55")),
		#("mlconfig/ada4s1.cfg", config.great3.path("ml", "%03i" % subfield, trainspname, "ada4s1_sum55/mini_ada4s1_sum55_2017-02-26T09-46-42")),
	]
	"""
	



"""
Running the new interface
"""

import numpy as np
import tenbilac


import logging
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


(inputs, targets) = tenbilac.utils.readpickle("data.pkl")

# And go:

ten = tenbilac.com.Tenbilac("tenbilac.cfg")
ten.train(inputs, targets)

# Demo of alternative way to load an existing tenbilac object:
#ten = tenbilac.com.Tenbilac("/vol/fohlen11/fohlen11_1/mtewes/tenbilac_demo_workdir/mini_tenbilac_2017-02-22T11-59-10")

preds = ten.predict(inputs)
print preds



"""
Classes to hold and manipulate data
"""

import numpy as np


import logging
logger = logging.getLogger(__name__)


class Normer:
	"""
	A Normer provides methods to "normalize" and later "denormalize" a numpy array, linearly rescaling it to be within 0 and 1 (type="01"),
	-1 and 1 (type="-11"), around 0 with a std of 1 (type="std"), or just factor-scaling it so that the maximum absolute value is 1 (type="sa1").
	
	For each feature, the same normalization is applied to all realizations of all cases.
	The details of the normalization are kept in the Normer object, so that one can simply use the same
	object to denorm stuff afterwards.
	
	This works with inputs (3D or 2D) and outputs or targets (2D), with indices
	(realization, feature, case) or (feature, case).
	
	All methods correctly work with masked arrays.
	
	"""

	def __init__(self, x, type="-11"):
		"""
		Does *not* normalize anything, just determines the normalization parameters!
		"""

		self.type = type
		
		if isinstance(x, np.ma.MaskedArray):
			logger.info("Building Normer of type '{}' with a masked array of shape {} and {} masked values".format(self.type, str(x.shape), np.sum(x.mask)))
			#np.ma.set_fill_value(x, 1.e20) # To notice things if the array gets filled by error.
		elif isinstance(x, np.ndarray):
			logger.info("Building Normer of type '{}' with an unmasked array of shape {}".format(self.type, x.shape))
		else:
			raise ValueError("x is not a numpy array")
		
		if x.ndim not in (2, 3):
			raise ValueError("Cannot handle this array shape")
			
		if x.shape[-2] > 100: # This is the number of features
			raise RuntimeError("Looks like you have more than 100 features or labels, this is probably a array format error !")
			
		if type in ["01", "-11"]:
			# For this, we need to compute the min and max value for every feature, along the realizations and cases.
		
			if x.ndim == 3:# If we have several realizations:
				mins = np.min(np.min(x, axis=0), axis=1) # Computes the min along the first and thrid axis.
				dists = np.max(np.max(x, axis=0), axis=1) - mins
				
				# All these np.min, np.max, np.mean, np.std work as expected also with masked arrays.
			elif x.ndim == 2:
				mins = np.min(x, axis=1) # Only along the second axes (i.e., "cases")
				dists = np.max(x, axis=1) - mins
							
			self.a = mins
			self.b = dists
		
		elif type == "sa1":
			# We only rescale the values so that the max amplitude is 1. This ensures that signs are kept.
			
			if x.ndim == 3:# If we have several realizations:
				scales = np.max(np.max(np.fabs(x), axis=0), axis=1) # max absolute value along the first and thrid axis.
			
			elif x.ndim == 2:
				scales = np.max(np.fabs(x), axis=1) # Only along the second axes (i.e., "cases")
				
			assert scales.ndim == 1 # Only the "feature" dimension remains.
			
			self.b = scales
			self.a = np.zeros(scales.shape)
			
					
		elif type == "std":
			
			if x.ndim == 3: # First using rollaxis to reshape array instead of using fancy reshape modes seemed safer...
				xreshaped = np.rollaxis(x, axis=1) # brings the feature as first index.
				xreshaped = np.reshape(xreshaped, (x.shape[1], -1)) # mixes reas and cases in second index.
				avgs = np.mean(xreshaped, axis=1) # along reas and cases
				stds = np.std(xreshaped, axis=1)
				
			elif x.ndim == 2: # Easy !
				avgs = np.mean(x, axis=1) # Along cases
				stds = np.std(x, axis=1)
						
			self.a = avgs
			self.b = stds
			
		else:
			raise RuntimeError("Unknown Normer type")		
		
		#logger.info(str(self))
		

	def __str__(self):
		return "Normer of type '{self.type}': a={self.a}, b={self.b}".format(self=self)


	def __call__(self, x):
		"""
		Returns the normalized data.
		"""
		
		logger.info("Normalizing array of shape {} with normer-type '{}'".format(x.shape, self.type))
		if x.ndim not in (2, 3):
			raise ValueError("Cannot handle this array shape")

		assert self.a.ndim == 1
		assert self.b.ndim == 1
		
		if x.shape[-2] != self.a.shape[0]:
			raise RuntimeError("Number of features does not match!")
			
		atiled = np.tile(self.a.reshape(self.a.size, 1), (1, x.shape[-1]))
		btiled = np.tile(self.b.reshape(self.b.size, 1), (1, x.shape[-1]))			
		res = (x - atiled) / btiled
					
		if self.type == "-11":
			res = 2.0*res - 1.0
		
		return res

	def denorm(self, x):
		"""
		Denorms the data
		"""
		
		if x.ndim not in (2, 3):
			raise ValueError("Cannot handle this array shape")

		assert self.a.ndim == 1
		assert self.b.ndim == 1
		
		if self.type == "-11":
			res = (x + 1.0) / 2.0
		else:
			res = x + 0.0

		if res.shape[-2] != self.a.shape[0]:
			raise RuntimeError("Number of features does not match!")
			
		atiled = np.tile(self.a.reshape(self.a.size, 1), (1, res.shape[-1]))
		btiled = np.tile(self.b.reshape(self.b.size, 1), (1, res.shape[-1]))			
		res = res * btiled + atiled
		
		return res



def demask(indata, no=1):
	"""
	Function that "splits" a potentially masked input 3D array into unmasked input and some appropriate mask
	that can be applied to the output.
	This allows us to write the neural network itself as if no data was masked, as long as the cost function
	is aware of the mask.
	
	The whole point: if any feature of a realization is maksed,
	the full realization should be disregarded.
	
	:param indata: 3D numpy array (rea, feature, case), typically input for training or prediction.
	:param no: The number of outputs of the network (only used to properly format the returned mask).
	
	:returns: a tuple (filleddata, outputsmask), where filledata has exactly the same shape as data,
		and outputsmask is 3D but with only "no" feature dimensions (rea, =no, case)
		If the input data is not masked, the returned outputsmask is "None".
	
	"""
	assert indata.ndim == 3
	
	if isinstance(indata, np.ma.MaskedArray):
			
		assert indata.mask.ndim == 3
		
		outputsmask = np.any(indata.mask, axis=1) # This is 2D (rea, gal)
		
		# Let's also compute a mask for galaxies, just to see how many are affected:
		galmask = np.any(outputsmask, axis=0) # This is 1D (gal)
		galmaskall = np.all(outputsmask, axis=0) # This is 1D (gal)
			
		txt = []
			
		txt.append("In total {0} realizations ({1:.2%}) will be disregarded due to {2} masked features.".format(np.sum(outputsmask), float(np.sum(outputsmask))/float(outputsmask.size), np.sum(indata.mask)))
		txt.append("This affects {0} ({1:.2%}) of the {2} cases,".format(np.sum(galmask), float(np.sum(galmask))/float(galmask.size), galmask.size))
		txt.append("and {0} ({1:.2%}) of the cases have no useable realizations at all.".format(np.sum(galmaskall), float(np.sum(galmaskall))/float(galmaskall.size)))
			
		logger.info(" ".join(txt))
			
		# Now we inflate this outputsmask to make it 3D (rea, label, gal)
		# Values are the same for all labels, but this is required for easy use in the error functions.
		outputsmask = np.swapaxes(np.tile(outputsmask, (no, 1, 1)), 0, 1)
			
		filleddata = indata.filled(fill_value=0.0) # Gives us a plain ndarray without mask.
		assert type(filleddata) == np.ndarray
			
	else:
		assert type(indata) == np.ndarray
		logger.info("The data has no mask, so nothing to demask...")
		filleddata = indata
		outputsmask = None
	
	return (filleddata, outputsmask)
	
	




class Traindata:
	"""
	A class to hold training data so that it can be efficiently used by Tenbilac.
	It has methods to set validation data, swap minibatches etc.
	
	
	Here we deal with avoiding masked arrays for the Net. Instead, we carry around boolean masks to be used by the error functions
	after the unmasked arrays were propagated through the network.
	This way, we can "run" the network without paying attention to the masks.
	Instead, we generate a mask for the ouputs, so that the errorfunction can disregard the masked realizations.
	All this masking stays the same for given training data, no need to compute this at every iteration...
	
	Naming conventions:
	fulltraininputs = the full training set
	valinputs = the full validation set
	traininputs = the current "mini batch" subset of the full training set (or the full training set, if no minibatch is set)
	
	
	The fulltrain... and val... stuff is set directly by the constructor.
	The train... stuff is set when a minibatch is created.
	
	"""



	def __init__(self, inputs, targets, auxinputs=None, valfrac=0.5, shuffle=True):
		"""
		
		:param inputs: masked 3D array with the inputs for the neural network, I'll take care of demasking it.
		:param targets: 2D array with the targets (should not be masked, as targets should all be known...)
		
		:param auxinputs: masked 3D array with auxiliary inputs, that is inputs that will be used by the error function,
			but not serve as inputs for the neural nets.
			As the auxinputs are not fed into a Net, we just keep a potential mask as it is.
		
		:param valfrac: Fraction of training data which should be used for the validation
	
		"""
		
		logger.info("Setting up the training data...")
		
		if inputs.ndim != 3 or targets.ndim != 2:
			raise ValueError("Sorry, for training I only accept 3D input and 2D targets.")
		if inputs.shape[2] != targets.shape[1] : # Checking for same number of cases
			raise ValueError("Shape of inputs {} not compatible with targets {}".format(inputs.shape, targets.shape))
		if type(targets) != np.ndarray:
			raise ValueError("Sorry, targets should not be masked")
	
		# We split the mask apart:
		(nomaskinputs, outputsmask) = demask(inputs, no=targets.shape[0])
		
		# Checking the auxinputs
		if auxinputs != None:
			if auxinputs.ndim != 3:
				raise ValueError("Please give me 3D auxinputs")
			if auxinputs.shape[2] != inputs.shape[2] or auxinputs.shape[0] != inputs.shape[0]: # Checking for same number of cases and realizations
				raise ValueError("Shape of auxinputs {} not compatible with inputs {}".format(auxinputs.shape, inputs.shape))
		
		

		# Now we cut away part of the data for validation purposes, and shuffle before doing so.
		
		ncases = inputs.shape[2]
		nvalcases = int(valfrac * ncases) # number of validation cases
		if nvalcases <= 0:
			raise RuntimeError("Please allow for some validation cases.")
		ntraincases = ncases - nvalcases

		if shuffle:
			logger.info("Shuffling training data and selecting {nvalcases} among {ncases} cases for validation...".format(ncases=ncases, nvalcases=nvalcases))
			caseindexes = np.arange(ncases)
			np.random.shuffle(caseindexes)
			trainindexes = caseindexes[0:ntraincases]
			valindexes = caseindexes[ntraincases:ncases]
			assert len(trainindexes) + len(valindexes) == ncases
			
			self.fulltraininputs = nomaskinputs[:,:,trainindexes]
			self.valinputs = nomaskinputs[:,:,valindexes]
			self.fulltraintargets = targets[:,trainindexes]
			self.valtargets = targets[:,valindexes]
			
			if outputsmask is None:
				self.fulltrainoutputsmask = None
				self.valoutputsmask = None
			else:
				self.fulltrainoutputsmask = outputsmask[:,:,trainindexes]
				self.valoutputsmask = outputsmask[:,:,valindexes]
			
			if auxinputs is None:
				self.fulltrainauxinputs = None
				self.valauxinputs = None
			else:
				self.fulltrainauxinputs = auxinputs[:,:,trainindexes]
				self.valauxinputs = auxinputs[:,:,valindexes]
			
				
			
		else: # then we just slice the arrays:
			logger.info("Selecting the last {nvalcases} among {ncases} cases for validation (no shuffling)".format(ncases=ncases, nvalcases=nvalcases))
			
			self.fulltraininputs = nomaskinputs[:,:,0:ntraincases]
			self.valinputs = nomaskinputs[:,:,ntraincases:ncases]
			self.fulltraintargets = targets[:,0:ntraincases]
			self.valtargets = targets[:,ntraincases:ncases]
			
			if outputsmask is None:
				self.fulltrainoutputsmask = None
				self.valoutputsmask = None
			else:
				self.fulltrainoutputsmask = outputsmask[:,:,0:ntraincases]
				self.valoutputsmask = outputsmask[:,:,ntraincases:ncases]
			
			if auxinputs is None:
				self.fulltrainauxinputs = None
				self.valauxinputs = None
			else:
				self.fulltrainauxinputs = auxinputs[:,:,0:ntraincases]
				self.valauxinputs = auxinputs[:,:,ntraincases:ncases]

			
			
		# Let's check that all this looks good:
		assert self.fulltraininputs.shape[2] == ntraincases
		assert self.valinputs.shape[2] == nvalcases			
		assert self.fulltraintargets.shape[1] == ntraincases
		assert self.valtargets.shape[1] == nvalcases			
		if outputsmask is not None:
			assert self.fulltrainoutputsmask.shape[2] == ntraincases
			assert self.valoutputsmask.shape[2] == nvalcases
		if auxinputs is not None:
			assert self.fulltrainauxinputs.shape[2] == ntraincases
			assert self.valauxinputs.shape[2] == nvalcases
					

		# By default we set the full training set as batch:
		self.fullbatch()


	def __str__(self):
		
		if self.getntrain() == self.getnfulltrain():
			ntraintxt = "{nfulltrain}".format(nfulltrain=self.getnfulltrain())	
		else:
			ntraintxt = "{nfulltrain}({ntrain})".format(nfulltrain=self.getnfulltrain(), ntrain=self.getntrain())	
		
		return "{nrea}*{ntraintxt}|{nval}".format(ntraintxt=ntraintxt, nval=self.getnval(), nrea=self.getnrea())



	def getnrea(self):
		"""
		The number of realizations
		"""
		return self.fulltraininputs.shape[0]

	def getni(self):
		"""
		The number of "features", i.e. input nodes of the ANN
		"""
		return self.fulltraininputs.shape[1]

	def getno(self):
		"""
		The number of labels, i.e. output nodes of the ANN
		"""
		return self.fulltraintargets.shape[0]

	def getntrain(self):
		"""
		Number of training cases in batch
		"""
		return self.traininputs.shape[2]
	
	def getnfulltrain(self):
		"""
		Total number of training cases (not validation cases!)
		"""
		return self.fulltraininputs.shape[2]

	def getnval(self):
		"""
		Number of validation cases
		"""
		return self.valinputs.shape[2]


	def fullbatch(self):
		"""
		Sets the full training sample as batch training data.
		"""
		logger.info("Setting the full training set to be used as batch.")
		self.traininputs = self.fulltraininputs
		self.trainoutputsmask = self.fulltrainoutputsmask # even if None, this works
		self.traintargets = self.fulltraintargets
		self.trainauxinputs = self.fulltrainauxinputs


	def random_minibatch(self, mbsize=None, mbfrac=0.1):
		"""
		Selects a random minibatch of the full training set
		:param mbsize: if given, I will select a minibatch of this size.
		:param mbfrac: same parameter, but expressed as fraction of the size of the fulltrain sample. Note that mbsize overwrites this, if given.
		"""
		
		if mbsize is None and mbfrac is None:
			self.fullbatch()
			return
			#raise RuntimeError("Please give a mbsize or mbfrac!")
		
		nfulltrain = self.getnfulltrain()
		
		if mbfrac != None:
			finalmbsize = int(mbfrac * nfulltrain)
		if mbsize != None: # If given, we overwrite the mbfrac computation!
			finalmbsize = mbsize
		
		
		if finalmbsize > nfulltrain:
			raise RuntimeError("Cannot select {finalmbsize} among {nfulltrain}".format(finalmbsize=finalmbsize, nfulltrain=nfulltrain))
		
		if finalmbsize <= 0:
			raise RuntimeError("Cannot select no sample!")
		
		logger.info("Randomly selecting new minibatch of {finalmbsize} (mbfrac={mbfrac}) among {nfulltrain} cases...".format(finalmbsize=finalmbsize, mbfrac=mbfrac, nfulltrain=nfulltrain))
		caseindexes = np.arange(nfulltrain)
		np.random.shuffle(caseindexes)
		caseindexes = caseindexes[0:finalmbsize]
			
		self.traininputs = self.fulltraininputs[:,:,caseindexes]
		self.traintargets = self.fulltraintargets[:,caseindexes]
		
		if self.fulltrainoutputsmask is not None:
			self.trainoutputsmask = self.fulltrainoutputsmask[:,:,caseindexes] # Yes, outputsmask is 3D
		else:
			self.trainoutputsmask = None
		
		if self.fulltrainauxinputs is not None:
			self.trainauxinputs = self.fulltrainauxinputs[:,:,caseindexes] # Also 3D
		else:
			self.trainauxinputs = None
	
	





import matplotlib
#matplotlib.use("AGG")
matplotlib.use("pdf")

import momentsml
import momentsml.tools
import momentsml.plot
from momentsml.tools.feature import Feature

import momentsmlgreat3
import astropy


import config
import numpy as np
import os
import matplotlib
import matplotlib.pyplot as plt
import matplotlib.ticker as plticker
from matplotlib.ticker import AutoMinorLocator, LogLocator, MultipleLocator, FixedLocator


from matplotlib import rc
rc('font',**{'family':'sans-serif','sans-serif':['Helvetica']})
## for Palatino and other serif fonts use:
#rc('font',**{'family':'serif','serif':['Palatino']})
rc('text', usetex=True)



import logging
logging.basicConfig(format=config.loggerformat, level=logging.DEBUG)
logger = logging.getLogger(__name__)

from matplotlib.ticker import Locator

#Fixing a minor locator symlog bug: 
#http://stackoverflow.com/questions/20470892/how-to-place-minor-ticks-on-symlog-scale
class MinorSymLogLocator(Locator):
	"""
	Dynamically find minor tick positions based on the positions of
	major ticks for a symlog scaling.
	"""
	def __init__(self, linthresh):
		"""
		Ticks will be placed between the major ticks.
		The placement is linear for x between -linthresh and linthresh,
		otherwise its logarithmically
		"""
		self.linthresh = linthresh

	def __call__(self):
		'Return the locations of the ticks'
		majorlocs = self.axis.get_majorticklocs()

		# iterate through minor locs
		minorlocs = []

		# handle the lowest part
		for i in xrange(1, len(majorlocs)):
			majorstep = majorlocs[i] - majorlocs[i-1]
			if abs(majorlocs[i-1] + majorstep/2) < self.linthresh:
				ndivs = 10
			else:
				ndivs = 9
			minorstep = majorstep / ndivs
			locs = np.arange(majorlocs[i-1], majorlocs[i], minorstep)[1:]
			minorlocs.extend(locs)

		return self.raise_if_exceeds(np.array(minorlocs))

	def tick_values(self, vmin, vmax):
		raise NotImplementedError('Cannot get tick locations for a '
								  '%s type.' % type(self))


catpath = config.great3.path("summary_{}.pkl".format(config.predcode))
#catpath = config.great3.path("{}_summary_{}.pkl".format(config.datasets["mimic-great3"], config.predcode))


cat = momentsml.tools.io.readpickle(catpath)


# Computing residuals

cat["pre_s1_res"] = cat["pre_s1"] - cat["tru_s1"]
cat["pre_s2_res"] = cat["pre_s2"] - cat["tru_s2"]

cat["pre_s1w_res"] = cat["pre_s1w"] - cat["tru_s1"]
cat["pre_s2w_res"] = cat["pre_s2w"] - cat["tru_s2"]



# Computing PSF rotated errors
# angles:
rotations = .5 * np.arctan2(cat["tru_psf_g2"], cat["tru_psf_g1"]) # Note the inverted order of components
		

# rotating  predictions:
cat["pre_s+"] = cat["pre_s1"] * np.cos(-2. * rotations) - cat["pre_s2"] * np.sin(-2. * rotations)
cat["pre_sx"] = cat["pre_s1"] * np.sin(-2. * rotations) + cat["pre_s2"] * np.cos(-2. * rotations)

# rotating weighted predictions:
cat["pre_s+w"] = cat["pre_s1w"] * np.cos(-2. * rotations) - cat["pre_s2w"] * np.sin(-2. * rotations)
cat["pre_sxw"] = cat["pre_s1w"] * np.sin(-2. * rotations) + cat["pre_s2w"] * np.cos(-2. * rotations)

# rotating truth:
cat["tru_s+"] = cat["tru_s1"] * np.cos(-2. * rotations) - cat["tru_s2"] * np.sin(-2. * rotations)
cat["tru_sx"] = cat["tru_s1"] * np.sin(-2. * rotations) + cat["tru_s2"] * np.cos(-2. * rotations)

# And again the residuals
cat["pre_s+_res"] = cat["pre_s+"] - cat["tru_s+"]
cat["pre_sx_res"] = cat["pre_sx"] - cat["tru_sx"]

cat["pre_s+w_res"] = cat["pre_s+w"] - cat["tru_s+"]
cat["pre_sxw_res"] = cat["pre_sxw"] - cat["tru_sx"]


import galsim

cat["tru_psf_e1"] = 0.0
cat["tru_psf_e2"] = 0.0

for row in cat:
	s = galsim.Shear(g1=row["tru_psf_g1"], g2=row["tru_psf_g2"])
	row["tru_psf_e1"] = s.e1
	row["tru_psf_e2"] = s.e2
	

"""
SMALL_SIZE = 14
MEDIUM_SIZE = 16
BIGGER_SIZE = 30

plt.rc('font', size=SMALL_SIZE)          # controls default text sizes
plt.rc('axes', titlesize=SMALL_SIZE)     # fontsize of the axes title
plt.rc('axes', labelsize=MEDIUM_SIZE)    # fontsize of the x and y labels
plt.rc('xtick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('ytick', labelsize=SMALL_SIZE)    # fontsize of the tick labels
plt.rc('legend', fontsize=SMALL_SIZE)    # legend fontsize
plt.rc('figure', titlesize=BIGGER_SIZE)  # fontsize of the figure title
#print momentsml.tools.table.info(cat)
#mets = momentsmlgreat3.utils.metrics(cat, ("tru_s1", "tru_s2"), ("pre_s1w", "pre_s2w"), psfgcols=("tru_psf_g1", "tru_psf_g2"))

"""
featc = Feature("psf_adamom_sigma", nicename=r"PSF $\mathtt{adamom\_sigma}$ [pix]")
cmap="plasma_r"


#fig = plt.figure(figsize=(8, 7))
#fig = plt.figure(figsize=(9, 4.5))
fig = plt.figure(figsize=(7.5, 6.4))

fig.text(.005, .94, config.great3.get_branchacronym().upper(), fontdict={"fontsize":22})



#ax = plt.axes([.09, paneltoplevel, panelxw, panelyw])

resr = 0.039
symthres = 0.004
alpha = 0.5
majorLocator = MultipleLocator(0.1)
minorlocator = MinorSymLogLocator(1.0)


#subs = [2.,3.,4.,5.,6.,7.,8.,9.,-2.,-3.,-4.,-5.,-6.,-7.,-8.,-9.]
#minorlocator = LogLocator(10,subs=subs[:])


tickpos = np.array([0.001, 0.002, 0.003, 0.004, 0.005, 0.006, 0.007, 0.008, 0.009, 0.01, 0.02, 0.03, 0.04, 0.05, 0.06, 0.07, 0.08, 0.09, 0.1])
minorlocator = FixedLocator(list(tickpos) + list(-tickpos))

ax = fig.add_subplot(2, 2, 1)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_e1", -0.2, 0.15, nicename=r"True $e_{1, \mathrm{PSF}}$"), Feature("pre_s1_res", -0.2, 0.2, nicename=r"Predicted $g_1$ $-$ True $g_1$"), featc,
cmap=cmap, showidline=True, yisres=True)
ax.set_yscale('symlog', linthreshy=symthres)
ax.yaxis.set_minor_locator(minorlocator)
ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
ax.set_title("Without weights")
ax.title.set_position([.5, 1.1])
ax.xaxis.set_major_locator(majorLocator)
#ax.yaxis.set_minor_locator(minorlocator)
# Hiding "0" as it collides with the 10-3:
#ax.axes.yaxis.set_ticklabels([" ", "-0.1", "-0.01", "-0.001"," ", "0.001", "0.01", "0.1"])
ticks = ax.yaxis.get_major_ticks()
ticks[4].label1.set_visible(False)

ax = fig.add_subplot(2, 2, 2)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_e1", -0.2, 0.15, nicename=r"True $e_{1, \mathrm{PSF}}$"), Feature("pre_s1w_res", -resr, resr, nicename=r"Predicted $g_1$ $-$ True $g_1$"), showidline=True, yisres=True,
alpha=alpha)
ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
ax.set_title("With weights")
ax.title.set_position([.5, 1.1])
ax.xaxis.set_major_locator(majorLocator)


ax = fig.add_subplot(2, 2, 3)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_e2", nicename=r"True $e_{2, \mathrm{PSF}}$"), Feature("pre_s2_res", -0.2, 0.2, nicename=r"Predicted $g_2$ $-$ True $g_2$"), featc,
cmap=cmap, showidline=True, yisres=True, hidecbar=True)
ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
ax.xaxis.set_major_locator(majorLocator)
ax.set_yscale('symlog', linthreshy=symthres)
ax.yaxis.set_minor_locator(minorlocator)
# Hiding "0" as it collides with the 10-3:
#ax.axes.yaxis.set_ticklabels([" ", "-0.1", "-0.01", "-0.001"," ", "0.001", "0.01", "0.1"])
ticks = ax.yaxis.get_major_ticks()
ticks[4].label1.set_visible(False)

ax = fig.add_subplot(2, 2, 4)
momentsml.plot.scatter.scatter(ax, cat, Feature("tru_psf_e2", nicename=r"True $e_{2, \mathrm{PSF}}$"), Feature("pre_s2w_res", -0.2, 0.2, nicename=r"Predicted $g_2$ $-$ True $g_2$"), featc,
cmap=cmap, showidline=True, yisres=True, hidecbar=True)
ax.set_yscale('symlog', linthreshy=symthres)
ax.yaxis.set_minor_locator(minorlocator)
ax.fill_between([-1, 1], -symthres, symthres, alpha=0.2, facecolor='darkgrey')
ax.xaxis.set_major_locator(majorLocator)
# Hiding "0" as it collides with the 10-3:
#ax.axes.yaxis.set_ticklabels([" ", "-0.1", "-0.01", "-0.001"," ", "0.001", "0.01", "0.1"])
ticks = ax.yaxis.get_major_ticks()
ticks[4].label1.set_visible(False)





plt.tight_layout()


plotpath = config.great3.path("fig_3_{}.pdf".format(config.great3.get_branchacronym()))
plt.savefig(plotpath)
#plt.show()
print plotpath



"""
Configuration for the GREAT3 scripts.
"""

import momentsmlgreat3 as mg3

loggerformat='\033[1;31m%(levelname)s\033[1;0m: %(name)s(%(funcName)s): \033[1;21m%(message)s\033[1;0m'


great3 = mg3.great3.GREAT3Run(
	experiment = "control",
	obstype = "ground",
	sheartype = "constant",
	datadir = "/vol/fohlen11/fohlen11_1/mtewes/GREAT3",
	truthdir = "/vol/fohlen11/fohlen11_1/mtewes/GREAT3/truth", # Only needed for final analysis plots
	workdir = "/vol/fohlen11/fohlen11_1/mtewes/2017_MomentsML_GREAT3/cgc_v2",
	g3publicdir = "/users/mtewes/code/great3-public",
	
	
	#subfields = [0,1,2],
	#subfields = [99],	# CGC best PSF subfields: [99, 156, 42],  worst : [2, 191, 166]

	#subfields = range(0, 200),
	
	#subfields = [1099], # Copy of 99, used for manual tests (to avoid messing up 99)

	#subfields = range(0, 50),
	#subfields = range(50, 100),
	#subfields = range(100, 150),
	#subfields = range(150, 200),

	#subfields = range(0, 100),
	#subfields = range(100, 200),

	ncpu = 20,
	skipdone = True
	)


### Script configuration ###


datasets = {
	#"train-shear":"ts-shear-nn-train-snc20",
	#"train-shear":"ts-shear-nn-train-snc100",
	#"train-shear":"ts-ell-n-train-rea100",
	"train-shear":"ts-ell-nn-train-rea10",
	
	"valid-shear":"vs-shear-n-G3-snc1000",
	
	#"train-weight":"tw-20c-2000r",
	#"train-weight":"tw-200c-5000r",
	#"train-weight":"tw-200c-500r",
	#"train-weight":"tw-200c-800r",
	#"train-weight":"tw-500c-200r",
	#"train-weight":"tw-100c-400r",
	#"train-weight":"tw-100c-1000r",
	"train-weight":"tw-200c-1000r",
	
	"valid-overall":"vo-200c-8000r",
	#"simobscompa":"simobscompa-G3",
	"simobscompa":"simobscompa-train",
	
	"mimic-great3":"sersicG3subfield"
	#"mimic-great3":"sersicG3subfield_nosnc"

}


shearconflist = [
	#("mlconfig/ada4s1.cfg", "mlconfig/sum55.cfg"), # Uncomment a line to run on only one component
	#("mlconfig/ada4s2.cfg", "mlconfig/sum55.cfg")

	("mlconfig/ada4g1.cfg", "mlconfig/sum55mass.cfg"), # Uncomment a line to run on only one component
	("mlconfig/ada4g2.cfg", "mlconfig/sum55mass.cfg")
	
	
]

weightconflist = [
	#("mlconfig/ada2s1w.cfg", "mlconfig/sum3w.cfg"),
	
	#("mlconfig/ada4s1w.cfg", "mlconfig/sum55w.cfg"),
	#("mlconfig/ada4s1w.cfg", "mlconfig/sum33w.cfg"),
	
	#("mlconfig/ada4s1w.cfg", "mlconfig/sum33wmass.cfg"), # predcode 1
	#("mlconfig/ada4s2w.cfg", "mlconfig/sum33wmass.cfg"),

	#("mlconfig/ada5s1w.cfg", "mlconfig/sum55wmass50.cfg"), # not much converged, predcode 2
	#("mlconfig/ada5s2w.cfg", "mlconfig/sum55wmass50.cfg"),


	("mlconfig/ada5s1w.cfg", "mlconfig/sum55wmass500.cfg"), # pre-training
	("mlconfig/ada5s2w.cfg", "mlconfig/sum55wmass500.cfg"),


	#("mlconfig/ada4s1w.cfg", "mlconfig/sum33w_nomb.cfg"),
	#("mlconfig/ada4s1w.cfg", "mlconfig/sum3w.cfg"),
]

#weightconflist = []

trainmode = "g" # When training for ellipticity, switch this to "g". It's for the plots.

predcode = "1" # String used in file names of GREAT3 predictions. Allows you to keep several predictions side by side.



import momentsml
import os
import config
import numpy as np

from momentsml.tools.feature import Feature

import matplotlib
import matplotlib.pyplot as plt



cat = momentsml.tools.io.readpickle(os.path.join(config.simmeasdir, config.datasets["si"], "groupmeascat.pkl"))

#print momentsml.tools.table.info(cat)


fig = plt.figure(figsize=(20, 13))


snr = Feature("snr", 0, 50, nicename="Measured S/N")
tru_rad = Feature("tru_rad", nicename="True half-light radius [pixel]")
tru_mag = Feature("tru_mag", 20, 26.5, nicename="Magnitude")
tru_sersicn = Feature("tru_sersicn", 0, 7, nicename="True Sersic index")

#snr = Feature("snr", nicename="Measured S/N")
#tru_rad = Feature("tru_rad", nicename="True half-light radius [pixel]")
#tru_mag = Feature("tru_mag", nicename="Magnitude")
#tru_sersicn = Feature("tru_sersicn", nicename="True Sersic index")


ax = fig.add_subplot(2, 3, 1)
momentsml.plot.scatter.scatter(ax, cat, tru_sersicn, tru_rad, sidehists=True, sidehistkwargs={"bins":20})
ax = fig.add_subplot(2, 3, 2)
momentsml.plot.scatter.scatter(ax, cat, tru_mag,  tru_rad, sidehists=True, sidehistkwargs={"bins":20})
ax = fig.add_subplot(2, 3, 3)
momentsml.plot.scatter.scatter(ax, cat, tru_mag,  tru_rad, snr)


ax = fig.add_subplot(2, 3, 4)
sel = momentsml.tools.table.Selector("draw", [
		("in", "tru_mag", 24.45, 24.55),
])
limcat = sel.select(cat)
momentsml.plot.scatter.scatter(ax, limcat, snr, tru_rad, sidehists=True, sidehistkwargs={"bins":20})
ax.grid()


ax = fig.add_subplot(2, 3, 5)
momentsml.plot.scatter.scatter(ax, cat, snr, tru_mag)
ax.grid()
ax = fig.add_subplot(2, 3, 6)
momentsml.plot.hist.hist(ax, cat, snr)




plt.tight_layout()
plt.show()
plt.close(fig) # Helps releasing memory when calling in large loops.





"""
High-level functionality to measure features on several images using multiprocessing.

We have two public functions:

- a onsims() which is specialised in exploring the file structure made by
  momentsml.sim.run.multi(), and run on them by calling
- general(), which can be used for any images.

The approach taken by general() is the following:

- make a list of _WorkerSetting objects, each of them describing the elementary task of
  measuring shapes on one single image.
- feed this list into _run(), which takes care of running the measurements.
  We use the simple multiprocessing.Pool to distribute the work on several cpus.
  This seems easier than implementing a pool-like queue ourself.

  
"""


import os
import sys
import glob
import datetime
import multiprocessing
import copy

from .. import tools

import logging
logger = logging.getLogger(__name__)


def onsims(simdir, simparams, measdir, measfct, measfctkwargs, ncpu=1, skipdone=True):
	"""
	Run the given measfct on sims created by sim.run.multi() for this simdir and simparams.
	This explores the simdir for potential images, and passes them to general().
	
	:param simdir: the simdir that you gave to sim.run.multi()
	:param simparams: the sim.params.Params that you gave to sim.run.multi()
		(Needed to get the simparams.name)
	:param measdir: Path to a directory where the catalogs should be written.
		As for the simdir, you can keep the same measdir for different simparams,
		as the results will be stored in subdirectories using simparams.name.
		But **the measdir should be changed if anything is changed on the measfct**.

	See the doc for general() for the explanation of the other parameters.
	
	As an illustration, the written file structure looks like::
	
		$ ls -1 measdir/name_of_simparams/
		20141017T185236_Cx9D5b_0_galimg_meascat.pkl
		20141017T185236_Cx9D5b_1_galimg_meascat.pkl
		20141017T185236_Cx9D5b_2_galimg_meascat.pkl
		20141017T185236_j8xBv6_0_galimg_meascat.pkl
		20141017T185236_j8xBv6_1_galimg_meascat.pkl
		20141017T185236_j8xBv6_2_galimg_meascat.pkl

	"""

	simworkdir = os.path.join(simdir, simparams.name)
	logger.info("Exploring the contents of simdir '%s'" % (simworkdir))
	
	incatfilepaths = sorted(glob.glob(os.path.join(simworkdir, "*_cat.pkl")))
	
	measworkdir = os.path.join(measdir, simparams.name)
	if not os.path.exists(measworkdir):
		os.makedirs(measworkdir)
	
	# The input lists for general():
	genincatfilepaths = [] # This is not incatfilepaths! The same catalogs will appear multiple times in here.
	genoutcatfilepaths = []
	genincatmetadicts = [] # Indeed we will have to set meta["img"] to point to the different realizations.
	
	
	for incatfilepath in incatfilepaths:
		
		# We read the catalog
		incat = tools.io.readpickle(incatfilepath)
		
		# And loop over the "declared" realization ImageInfo objects:
		for imgrea in incat.meta["imgreas"]:
			
			# Let's check that the declared image file does exist:
			if not os.path.exists(imgrea.filepath):
				logger.warning("Could not find image realization '%s', will skip it" % (imgrea.filepath))
				continue
			
			# If everything seems ok, we prepare the entries for general for this realization
			outcatfilepath = os.path.join(measdir, simparams.name, imgrea.name + "_meascat.pkl")
			
			# Now the ImageInfo object for meta["img"]. We make a new object, based on the
			# info for this realization, and add a workdir (in case measfct uses it).
			workdir = os.path.join(measdir, simparams.name, imgrea.name + "_workdir")
			imageinfo = tools.imageinfo.ImageInfo(
				imgrea.filepath, imgrea.xname, imgrea.yname, imgrea.stampsize,
				workdir = workdir
				)
			
			genincatfilepaths.append(incatfilepath)
			genoutcatfilepaths.append(outcatfilepath)
			genincatmetadicts.append({"img":imageinfo})
				
	
	# And we pass the ball to general():
	general(genincatfilepaths, genoutcatfilepaths, measfct, measfctkwargs,
		ncpu=ncpu, skipdone=skipdone, incatmetadicts=genincatmetadicts)
	


def general(incatfilepaths, outcatfilepaths, measfct, measfctkwargs, ncpu=1, skipdone=True, incatmetadicts=None):
	"""
	Run the given shape measurement (measfct) on your images using a multiprocessing pool map (ncpu).
	This measfct must be MomentsML-compliant, i.e. find all the required images by reading the meta of the passed catalog.
	Note that if you want to run on MomentsML simulations, see the function onsims() above (which calls this function).
	
	
	:param incatfilepaths: list of paths to the pickled input catalogs.
		These catalogs must contain the information required by measfct (and the measfctkwargs)
		to run on the image file(s) specified as ImageInfo objects in the catalog's meta dicts.
		Note that there is a trick to run with **different meta dicts**, see argument incatmetadicts below.
	:param outcatfilepaths: list of paths where the corresponding output catalog pickles should be written.
	:param measfct: The function that measures the sources. It has to be MomentsML-compliant, i.e., it has to find
		all the required images by reading the meta of the passed catalog.
		Often you will define this funtion yourself, as you might want to run different algorithms in one shot.
	:param measfctkwargs: keyword arguments controlling the behavior of the measfct.
		Usually these stay the same for every catalog/image (then give only one dict), but sometimes you might
		want to explore a range of different settings (then give a list of dicts, as long as incatfilepaths).
	:type measfctkwargs: dict or list of dicts
	:param ncpu: Maximum number of processes that should be used. Default is 1.
		Set to 0 for maximum number of available CPUs.
	:type ncpu: int
	:type skipdone: If set to False, re-run even if the output catalog already exists.
	:param incatmetadicts: If specified, these dicts will update the meta of the incats (for example
		the meta["img"] ...)
		This trick is very useful when running over several images of the same incat. Using this trick,
		you don't have to prepare and write to disk one copy of incat for every image.
	:type incatmetadicts: list of dicts


	.. warning:: If called "in parallel" (e.g., from several python scripts launched at about the same time),
		(in the worst case) all the work in every script might be run, i.e., skipdone will not work.
		Not a safety issue (hence currently no fix for this).  We can see if we want a fix for that.
		(can be done e.g. with temporary files written/detected and deleted by the workers)

	"""
	
	# Some trivial tests:
	
	n = len(incatfilepaths)
	if len(outcatfilepaths) != n:
		raise RuntimeError("The input/output catalog lists must have the same length")
	if incatmetadicts is not None:
		if len(incatmetadicts) != n:
			raise RuntimeError("The list incatmetadicts does not have the same length as incatfilepaths")
	else: # We turn this None into a list of None:
		incatmetadicts = [None] * n
		
		
	# If measfctkwargs is a simple dict, we make a list of identical dicts.
	
	if type(measfctkwargs) is dict:
		measfctkwargslist = [copy.deepcopy(measfctkwargs) for i in range(n)]
	else:
		measfctkwargslist = measfctkwargs
		if len(measfctkwargs) != n:
			raise RuntimeError("Your list of measfctkwargs does not have the same length as incatfilepaths")
			
	# It could be that the measfct wants to use the workdir of each image, if such workdirs are set.
	# In this case, it would be nice to test if all the workdirs are different (as they probably should).
	# A priori it seems not necessery to enforce this, but it's a simple way to avoid race conditions
	# at directory creation, or any unwanted file overwriting.
	# However, for this we would have to open the incats from the main process, this seems weird.
	# And so we skip this test.
	
	# Now we prepare the parallel processing.
	wslist = [] # The list to be filled with workersettings
	
	for (incatfilepath, outcatfilepath, measfctkwargsdict, incatmetadict) in zip(incatfilepaths, outcatfilepaths, measfctkwargslist, incatmetadicts):
			
		if skipdone and os.path.exists(outcatfilepath):
			logger.info("Output catalog %s already exists, skipping this one..." % (outcatfilepath))	
			continue
						
		ws = _WorkerSettings(incatfilepath, outcatfilepath, measfct, measfctkwargsdict, incatmetadict)
		wslist.append(ws)
	
	logger.info("Ready to run measurements on %i images." % (len(wslist)))
			
	# And we run a pool of workers on this wslist.
	_run(wslist, ncpu)
		



class _WorkerSettings():
	"""
	A class that holds together all the settings for measuring an image.
	"""
	
	def __init__(self, incatfilepath, outcatfilepath, measfct, measfctkwargs, incatmetadict):
		
		self.incatfilepath = incatfilepath
		self.outcatfilepath = outcatfilepath
		self.measfct = measfct
		self.measfctkwargs = measfctkwargs
		self.incatmetadict = incatmetadict
	
	def __str__(self):
		return "%s" % (os.path.basename(self.incatfilepath))



def _worker(ws):
	"""
	Worker function that the different processes will execute, processing the
	_WorkerSettings objects.
	"""
	starttime = datetime.datetime.now()
	p = multiprocessing.current_process()
	logger.info("%s is starting to measure catalog %s with PID %s" % (p.name, str(ws), p.pid))
	logger.debug("%s uses measfctkwargs %s" % (p.name, str(ws.measfctkwargs)))
	
	# Read input catalog
	incat = tools.io.readpickle(ws.incatfilepath)
	if ws.incatmetadict is not None:
		logger.debug("%s updates the catalog meta dict with %s" % (p.name, str(ws.incatmetadict)))
		incat.meta.update(ws.incatmetadict)
	
	
	# Run measfct, it will read the image by itself:
	logger.debug("%s will now run on image %s" % (p.name, incat.meta["img"].name))
	outcat = ws.measfct(incat, **ws.measfctkwargs)
	
	# Write output catalog
	tools.io.writepickle(outcat, ws.outcatfilepath)

	endtime = datetime.datetime.now()
	logger.info("%s is done, it took %s" % (p.name, str(endtime - starttime)))


def _run(wslist, ncpu):
	"""
	Wrapper around multiprocessing.Pool with some verbosity.
	"""
	
	if len(wslist) == 0: # This test is useful, as pool.map otherwise starts and is a pain to kill.
		logger.info("No images to measure.")
		return

	# We do not want to see the log from the low-level stuff:
	mutemodules = ["momentsml.meas.galsim_adamom", "momentsml.meas.sewfunc",
			"sewpy.sewpy", "momentsml.tools.io", "momentsml.tools.image"]
	for modulename in mutemodules:
		lowlevellogger = logging.getLogger(modulename)
		#lowlevellogger.propagate = False
		#lowlevellogger.setLevel(logging.WARNING)


	if ncpu == 0:
		try:
			ncpu = multiprocessing.cpu_count()
		except:
			logger.warning("multiprocessing.cpu_count() is not implemented!")
			ncpu = 1
	
	starttime = datetime.datetime.now()
	
	logger.info("Starting the measurement on %i images using %i CPUs" % (len(wslist), ncpu))
	
	if ncpu == 1: # The single process way (MUCH MUCH EASIER TO DEBUG...)
		map(_worker, wslist)
	
	else:
		pool = multiprocessing.Pool(processes=ncpu)
		pool.map(_worker, wslist)
		pool.close()
		pool.join()
	
	endtime = datetime.datetime.now()
	logger.info("Done, the total measurement time was %s" % (str(endtime - starttime)))
	
	# We unmute the mutemodules:
	#for modulename in mutemodules:
	#	lowlevellogger = logging.getLogger(modulename)
	#	lowlevellogger.propagate = True






"""
Subpackage for genrating plots

The modules in this subpackage typically define functions that plot something on matplotlib axes passed as first argument, based on the content of some astropy tables.
"""
__all__ = []

import utils
import scatter
import hist
import hexbin
import contour
import bin
import mcbin
import figures
import exp





"""
I/O functions directly related to the GREAT3 files
"""

import logging
import numpy as np
import os

import astropy.table

logger = logging.getLogger(__name__)



def readgalcat(branch, subfield, xt=None, yt=None):
	"""
	I read the galaxy catalog for a given branch and subfield.
	:returns: An astropy.table 
	"""
	
	filepath = branch.galcatfilepath(subfield, xt, yt)
	
	# There is probably a direct astropy way to read the data
	#cat = astropy.table.Table.read(filepath, format="ascii")
	#cat.rename_column('col1', 'ID')
	
	# But instead, let's try with numpy to illustrate maximum control:
	data = np.loadtxt(filepath)
	
	if xt is None and yt is None:
		assert data.shape[1] == 3
	else:
		assert data.shape[1] == 5
	
	ids = [int(line[2]) for line in data]
	xs = data[:,0]+1.5 # To get the same pixel convention as MomentsML
	ys = data[:,1]+1.5 # idem
	
	if xt is None and yt is None:
		cat = astropy.table.Table([ids, xs, ys],
				names=('ID', 'x', 'y'),
				meta = {"branch":branch, "subfield":subfield, "filepath":filepath}
				)
	else:
		cat = astropy.table.Table([ids, xs, ys, data[:,3], data[:,4]],
				names=('ID', 'x', 'y', 'tile_x_pos_deg', 'tile_y_pos_deg'),
				meta = {"branch":branch, "subfield":subfield, 
					"filepath":filepath, "xt":xt, 'yt':yt}
				)
		
	logger.info("Read %i sources from %s" % (len(cat), filepath))

	return cat


def readshear(filepath):
	"""
	Reads the "shear_params-000.txt" files and returns (g1, g2)
	"""
	logger.info("Reading shear from %s..." % (filepath))

	data = np.loadtxt(filepath)
	return data
	



"""
This is tenbilac:
https://github.com/mtewes/tenbilac
"""

from . import com
from . import utils
from . import net
from . import wnet
from . import multnet
from . import layer
from . import err
from . import regul
from . import act
from . import data
from . import train
from . import plot
from . import committee
from . import opt
from . import parmap
